import copy
import itertools
import multiprocessing as mp
import numbers
import platform
import sys
import warnings

import casadi
import numpy as np

import pybamm
from pybamm import ParameterValues
from pybamm.expression_tree.binary_operators import _Heaviside


class BaseSolver:
    """Solve a discretised model.

    Parameters
    ----------
    method : str, optional
        The method to use for integration, specific to each solver
    rtol : float, optional
        The relative tolerance for the solver (default is 1e-6).
    atol : float, optional
        The absolute tolerance for the solver (default is 1e-6).
    root_method : str or pybamm algebraic solver class, optional
        The method to use to find initial conditions (for DAE solvers).
        If a solver class, must be an algebraic solver class.
        If "casadi",
        the solver uses casadi's Newton rootfinding algorithm to find initial
        conditions. Otherwise, the solver uses 'scipy.optimize.root' with method
        specified by 'root_method' (e.g. "lm", "hybr", ...)
    root_tol : float, optional
        The tolerance for the initial-condition solver (default is 1e-6).
    extrap_tol : float, optional
        The tolerance to assert whether extrapolation occurs or not. Default is 0.
    output_variables : list[str], optional
        List of variables to calculate and return. If none are specified then
        the complete state vector is returned (can be very large) (default is [])
    on_extrapolation : str, optional
        What to do if the solver is extrapolating. Options are "warn", "error", or "ignore".
        Default is "warn".
    on_failure : str, optional
        What to do if a solver error flag occurs. Options are "warn", "error", or "ignore".
        Default is "raise".
    """

    def __init__(
        self,
        method=None,
        rtol=1e-6,
        atol=1e-6,
        root_method=None,
        root_tol=1e-6,
        extrap_tol=None,
        on_extrapolation=None,
        on_failure=None,
        output_variables=None,
    ):
        self.method = method
        self.rtol = rtol
        self.atol = atol
        self.root_tol = root_tol
        self.root_method = root_method
        self.extrap_tol = extrap_tol or -1e-10
        self.output_variables = [] if output_variables is None else output_variables
        self._on_extrapolation = on_extrapolation or "warn"
        self._on_failure = on_failure or "raise"
        self._model_set_up = {}

        # Defaults, can be overwritten by specific solver
        self.name = "Base solver"
        self._ode_solver = False
        self._algebraic_solver = False
        self._supports_interp = False
        self._supports_t_eval_discontinuities = False
        self.computed_var_fcns = {}
        self._mp_context = self.get_platform_context(platform.system())

    @property
    def ode_solver(self):
        return self._ode_solver

    @property
    def algebraic_solver(self):
        return self._algebraic_solver

    @property
    def supports_interp(self):
        return self._supports_interp

    @property
    def supports_t_eval_discontinuities(self):
        return self._supports_t_eval_discontinuities

    @property
    def supports_parallel_solve(self):
        return False

    @property
    def on_extrapolation(self):
        return self._on_extrapolation

    @on_extrapolation.setter
    def on_extrapolation(self, value):
        if value not in ["warn", "error", "ignore"]:
            raise ValueError("on_extrapolation must be 'warn', 'raise', or 'ignore'")
        self._on_extrapolation = value

    @property
    def on_failure(self):
        return self._on_failure

    @on_failure.setter
    def on_failure(self, value):
        if value not in ["warn", "error", "ignore"]:
            raise ValueError("on_failure must be 'warn', 'raise', or 'ignore'")
        self._on_failure = value

    @property
    def root_method(self):
        return self._root_method

    @root_method.setter
    def root_method(self, method):
        if method == "casadi":
            method = pybamm.CasadiAlgebraicSolver(self.root_tol)
        elif isinstance(method, str):
            method = pybamm.AlgebraicSolver(method, self.root_tol)
        elif not (
            method is None
            or (
                isinstance(method, pybamm.BaseSolver)
                and method.algebraic_solver is True
            )
        ):
            raise pybamm.SolverError("Root method must be an algebraic solver")
        self._root_method = method

    def copy(self):
        """Returns a copy of the solver"""
        new_solver = copy.copy(self)
        # clear _model_set_up
        new_solver._model_set_up = {}
        return new_solver

    def set_up(self, model, inputs=None, t_eval=None, ics_only=False):
        """Unpack model, perform checks, and calculate jacobian.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate. Must have attributes rhs and
            initial_conditions
        inputs : dict, optional
            Any input parameters to pass to the model when solving
        t_eval : numeric type, optional
            The times at which to stop the integration due to a discontinuity in time.
        """
        inputs = inputs or {}

        if ics_only:
            pybamm.logger.info("Start solver set-up, initial_conditions only")
        else:
            pybamm.logger.info("Start solver set-up")

        self._check_and_prepare_model_inplace(model, inputs, ics_only)

        # set default calculate sensitivities on model
        if not hasattr(model, "calculate_sensitivities"):
            model.calculate_sensitivities = []

        self._set_up_model_sensitivities_inplace(model, inputs)

        vars_for_processing = self._get_vars_for_processing(model, inputs)

        # Process initial conditions
        initial_conditions, _, jacp_ic, _ = process(
            model.concatenated_initial_conditions,
            "initial_conditions",
            vars_for_processing,
            use_jacobian=False,
        )
        model.initial_conditions_eval = initial_conditions
        model.jacp_initial_conditions_eval = jacp_ic

        # set initial conditions
        self._set_initial_conditions(model, 0.0, inputs)

        if ics_only:
            pybamm.logger.info("Finish solver set-up")
            return

        # Process rhs, algebraic, residual and event expressions
        # and wrap in callables
        rhs, jac_rhs, jacp_rhs, jac_rhs_action = process(
            model.concatenated_rhs, "RHS", vars_for_processing
        )

        algebraic, jac_algebraic, jacp_algebraic, jac_algebraic_action = process(
            model.concatenated_algebraic, "algebraic", vars_for_processing
        )

        # combine rhs and algebraic functions
        if len(model.rhs) == 0:
            rhs_algebraic = model.concatenated_algebraic
        elif len(model.algebraic) == 0:
            rhs_algebraic = model.concatenated_rhs
        else:
            rhs_algebraic = pybamm.NumpyConcatenation(
                model.concatenated_rhs, model.concatenated_algebraic
            )

        (
            rhs_algebraic,
            jac_rhs_algebraic,
            jacp_rhs_algebraic,
            jac_rhs_algebraic_action,
        ) = process(rhs_algebraic, "rhs_algebraic", vars_for_processing)

        (
            casadi_switch_events,
            terminate_events,
            interpolant_extrapolation_events,
            t_discon_constant,
            discontinuity_events,
        ) = self._set_up_events(model, t_eval, inputs, vars_for_processing)

        # Add the solver attributes
        model.rhs_eval = rhs
        model.algebraic_eval = algebraic
        model.rhs_algebraic_eval = rhs_algebraic

        model.terminate_events_eval = terminate_events
        model.interpolant_extrapolation_events_eval = interpolant_extrapolation_events
        model.discontinuity_events_eval = discontinuity_events
        model.t_discon_constant = t_discon_constant

        model.jac_rhs_eval = jac_rhs
        model.jac_rhs_action_eval = jac_rhs_action
        model.jacp_rhs_eval = jacp_rhs

        model.jac_algebraic_eval = jac_algebraic
        model.jac_algebraic_action_eval = jac_algebraic_action
        model.jacp_algebraic_eval = jacp_algebraic

        model.jac_rhs_algebraic_eval = jac_rhs_algebraic
        model.jac_rhs_algebraic_action_eval = jac_rhs_algebraic_action
        model.jacp_rhs_algebraic_eval = jacp_rhs_algebraic

        # Save CasADi functions for the CasADi solver
        # Save CasADi functions for solvers that use CasADi
        # Note: when we pass to casadi the ode part of the problem must be in
        if isinstance(self.root_method, pybamm.CasadiAlgebraicSolver) or isinstance(
            self, pybamm.CasadiSolver | pybamm.CasadiAlgebraicSolver
        ):
            # can use DAE solver to solve model with algebraic equations only
            if len(model.rhs) > 0:
                t_casadi = vars_for_processing["t_casadi"]
                y_casadi = vars_for_processing["y_casadi"]
                p_casadi_stacked = vars_for_processing["p_casadi_stacked"]
                mass_matrix_inv = casadi.MX(model.mass_matrix_inv.entries)
                explicit_rhs = mass_matrix_inv @ rhs(
                    t_casadi, y_casadi, p_casadi_stacked
                )
                model.casadi_rhs = casadi.Function(
                    "rhs", [t_casadi, y_casadi, p_casadi_stacked], [explicit_rhs]
                )
            model.casadi_switch_events = casadi_switch_events
            model.casadi_algebraic = algebraic
            model.casadi_sensitivities = jacp_rhs_algebraic
            model.casadi_sensitivities_rhs = jacp_rhs
            model.casadi_sensitivities_algebraic = jacp_algebraic

        if getattr(self.root_method, "algebraic_solver", False):
            self.root_method.set_up_root_solver(model, inputs, t_eval)

        # if output_variables specified then convert functions to casadi
        # expressions for evaluation within the respective solver
        self.computed_var_fcns = {}
        self.computed_dvar_dy_fcns = {}
        self.computed_dvar_dp_fcns = {}
        self._time_integral_vars = {}
        for key in self.output_variables:
            # Check for any ExplicitTimeIntegral or DiscreteTimeSum variables
            processed_time_integral = (
                pybamm.ProcessedVariableTimeIntegral.from_pybamm_var(
                    model.variables_and_events[key],
                    model.len_rhs_and_alg,
                )
            )
            # We will evaluate the sum node in the solver and sum it afterwards
            if processed_time_integral is None:
                var = model.variables_and_events[key]
            else:
                var = processed_time_integral.sum_node
                self._time_integral_vars[key] = processed_time_integral

            # Generate Casadi function to calculate variable and derivates
            # to enable sensitivites to be computed within the solver
            (
                self.computed_var_fcns[key],
                self.computed_dvar_dy_fcns[key],
                self.computed_dvar_dp_fcns[key],
                _,
            ) = process(
                var,
                BaseSolver._wrangle_name(key),
                vars_for_processing,
                use_jacobian=True,
                return_jacp_stacked=True,
            )

        pybamm.logger.info("Finish solver set-up")

    def _set_initial_conditions(self, model, time, inputs):
        # model should have been discretised or an error raised in Self._check_and_prepare_model_inplace
        len_tot = model.len_rhs_and_alg
        y_zero = np.zeros((len_tot, 1))

        casadi_format = model.convert_to_format == "casadi"
        if casadi_format:
            # stack inputs
            inputs_y0_ics = casadi.vertcat(*[x for x in inputs.values()])
        else:
            inputs_y0_ics = inputs

        model.y0 = model.initial_conditions_eval(time, y_zero, inputs_y0_ics)

        if model.jacp_initial_conditions_eval is None:
            model.y0S = None
            return

        if casadi_format:
            inputs_jacp_ics = inputs_y0_ics
        else:
            # we are calculating the derivative wrt the inputs
            # so need to make sure we convert int -> float
            # This is to satisfy JAX jacfwd function which requires
            # float inputs
            inputs_jacp_ics = {
                key: float(value) if isinstance(value, int) else value
                for key, value in inputs.items()
            }

        model.y0S = model.jacp_initial_conditions_eval(time, y_zero, inputs_jacp_ics)

    @classmethod
    def _wrangle_name(cls, name: str) -> str:
        """
        Wrangle a function name to replace special characters
        """
        replacements = [
            (" ", "_"),
            ("[", ""),
            ("]", ""),
            (".", "_"),
            ("-", "_"),
            ("(", ""),
            (")", ""),
            ("%", "prc"),
            (",", ""),
            (".", ""),
        ]
        name = "v_" + name.casefold()
        for string, replacement in replacements:
            name = name.replace(string, replacement)
        return name

    def _check_and_prepare_model_inplace(self, model, inputs, ics_only):
        """
        Performs checks on the model and prepares it for solving.
        """
        # Check model.algebraic for ode solvers
        if self.ode_solver is True and len(model.algebraic) > 0:
            raise pybamm.SolverError(
                f"Cannot use ODE solver '{self.name}' to solve DAE model"
            )
        # Check model.rhs for algebraic solvers
        if self._algebraic_solver is True and len(model.rhs) > 0:
            raise pybamm.SolverError(
                """Cannot use algebraic solver to solve model with time derivatives"""
            )
        # casadi solver won't allow solving algebraic model so we have to raise an
        # error here
        if isinstance(self, pybamm.CasadiSolver) and len(model.rhs) == 0:
            raise pybamm.SolverError(
                "Cannot use CasadiSolver to solve algebraic model, "
                "use CasadiAlgebraicSolver instead"
            )
        # Discretise model if it isn't already discretised
        # This only works with purely 0D models, as otherwise the mesh and spatial
        # method should be specified by the user
        if model.is_discretised is False:
            try:
                disc = pybamm.Discretisation()
                disc.process_model(model)
            except pybamm.DiscretisationError as e:
                raise pybamm.DiscretisationError(
                    "Cannot automatically discretise model, "
                    f"model should be discretised before solving ({e})"
                ) from e

        if (
            isinstance(self, pybamm.CasadiSolver | pybamm.CasadiAlgebraicSolver)
        ) and model.convert_to_format != "casadi":
            pybamm.logger.warning(
                f"Converting {model.name} to CasADi for solving with CasADi solver"
            )
            model.convert_to_format = "casadi"
        if (
            isinstance(self.root_method, pybamm.CasadiAlgebraicSolver)
            and model.convert_to_format != "casadi"
        ):
            pybamm.logger.warning(
                f"Converting {model.name} to CasADi for calculating ICs with CasADi"
            )
            model.convert_to_format = "casadi"

    @staticmethod
    def _get_vars_for_processing(model, inputs):
        vars_for_processing = {
            "model": model,
        }

        if model.convert_to_format != "casadi":
            # Create Jacobian from concatenated rhs and algebraic
            y = pybamm.StateVector(slice(0, model.len_rhs_and_alg))
            # set up Jacobian object, for re-use of dict
            jacobian = pybamm.Jacobian()
            vars_for_processing.update({"y": y, "jacobian": jacobian})
            return vars_for_processing

        else:
            # Convert model attributes to casadi
            t_casadi = casadi.MX.sym("t")
            # Create the symbolic state vectors
            y_diff = casadi.MX.sym("y_diff", model.len_rhs)
            y_alg = casadi.MX.sym("y_alg", model.len_alg)
            y_casadi = casadi.vertcat(y_diff, y_alg)
            p_casadi = {}
            for name, value in inputs.items():
                if isinstance(value, numbers.Number):
                    p_casadi[name] = casadi.MX.sym(name)
                else:
                    p_casadi[name] = casadi.MX.sym(name, value.shape[0])
            p_casadi_stacked = casadi.vertcat(*[p for p in p_casadi.values()])
            vars_for_processing.update(
                {
                    "t_casadi": t_casadi,
                    "y_diff": y_diff,
                    "y_alg": y_alg,
                    "y_casadi": y_casadi,
                    "p_casadi": p_casadi,
                    "p_casadi_stacked": p_casadi_stacked,
                }
            )

            return vars_for_processing

    @staticmethod
    def _set_up_model_sensitivities_inplace(model, inputs):
        """
        Set up model attributes related to sensitivities.
        """
        has_mass_matrix = model.mass_matrix is not None
        has_mass_matrix_inv = model.mass_matrix_inv is not None

        if not has_mass_matrix:
            return

        model.mass_matrix = pybamm.Matrix(
            model.mass_matrix.entries[: model.len_rhs_and_alg, : model.len_rhs_and_alg]
        )
        if has_mass_matrix_inv:
            model.mass_matrix_inv = pybamm.Matrix(
                model.mass_matrix_inv.entries[: model.len_rhs, : model.len_rhs]
            )

    def _set_up_events(self, model, t_eval, inputs, vars_for_processing):
        # Check for heaviside and modulo functions in rhs and algebraic and add
        # discontinuity events if these exist.
        # Note: only checks for the case of t < X, t <= X, X < t, or X <= t,
        # but also accounts for the fact that t might be dimensional
        tf = np.max(t_eval)

        def supports_t_eval_discontinuities(expr):
            # Only IDAKLUSolver supports discontinuities represented by t_eval
            return self.supports_t_eval_discontinuities and expr.is_constant()

        # Find all the constant time-based discontinuities
        t_discon = []

        def append_t_discon(t):
            t_discon.append(t)

        def heaviside_event(symbol, expr):
            model.events.append(
                pybamm.Event(
                    str(symbol),
                    expr,
                    pybamm.EventType.DISCONTINUITY,
                )
            )

        def heaviside_t_discon(symbol, expr):
            value = expr.evaluate(0, model.y0.full(), inputs=inputs)
            append_t_discon(value)

            if isinstance(symbol, pybamm.EqualHeaviside):
                if symbol.left == pybamm.t:
                    # t <= x
                    # Stop at t = x and right after t = x
                    append_t_discon(np.nextafter(value, np.inf))
                else:
                    # t >= x
                    # Stop at t = x and right before t = x
                    append_t_discon(np.nextafter(value, -np.inf))
            elif isinstance(symbol, pybamm.NotEqualHeaviside):
                if symbol.left == pybamm.t:
                    # t < x
                    # Stop at t = x and right before t = x
                    append_t_discon(np.nextafter(value, -np.inf))
                else:
                    # t > x
                    # Stop at t = x and right after t = x
                    append_t_discon(np.nextafter(value, np.inf))
            else:
                raise ValueError(
                    f"Unknown heaviside function: {symbol}"
                )  # pragma: no cover

        def modulo_event(symbol, expr, num_events):
            for i in np.arange(num_events):
                model.events.append(
                    pybamm.Event(
                        str(symbol),
                        expr * pybamm.Scalar(i + 1),
                        pybamm.EventType.DISCONTINUITY,
                    )
                )

        def modulo_t_discon(symbol, expr, num_events):
            value = expr.evaluate(0, model.y0.full(), inputs=inputs)
            for i in np.arange(num_events):
                t = value * (i + 1)
                # Stop right before t and at t
                append_t_discon(np.nextafter(t, -np.inf))
                append_t_discon(t)

        for symbol in itertools.chain(
            model.concatenated_rhs.pre_order(),
            model.concatenated_algebraic.pre_order(),
        ):
            if isinstance(symbol, _Heaviside):
                if symbol.right == pybamm.t:
                    expr = symbol.left
                elif symbol.left == pybamm.t:
                    expr = symbol.right
                else:
                    # Heaviside function does not contain pybamm.t as an argument.
                    # Do not create an event
                    continue  # pragma: no cover

                if supports_t_eval_discontinuities(expr):
                    heaviside_t_discon(symbol, expr)
                else:
                    heaviside_event(symbol, expr)

            elif isinstance(symbol, pybamm.Modulo) and symbol.left == pybamm.t:
                expr = symbol.right
                num_events = 200 if (t_eval is None) else (tf // expr.value)

                if supports_t_eval_discontinuities(expr):
                    modulo_t_discon(symbol, expr, num_events)
                else:
                    modulo_event(symbol, expr, num_events)
            else:
                continue

        casadi_switch_events = []
        terminate_events = []
        interpolant_extrapolation_events = []
        discontinuity_events = []
        for n, event in enumerate(model.events):
            if event.event_type == pybamm.EventType.DISCONTINUITY:
                # discontinuity events are evaluated before the solver is called,
                # so don't need to process them
                discontinuity_events.append(event)
            elif event.event_type == pybamm.EventType.SWITCH and (
                isinstance(self, pybamm.CasadiSolver)
                and self.mode == "fast with events"
                and model.algebraic != {}
            ):
                # Save some events to casadi_switch_events for the 'fast with
                # events' mode of the casadi solver
                # We only need to do this if the model is a DAE model
                # see #1082
                k = 20
                # address numpy 1.25 deprecation warning: array should have
                # ndim=0 before conversion
                init_sign = float(
                    np.sign(event.evaluate(0, model.y0.full(), inputs=inputs)).item()
                )
                # We create a sigmoid for each event which will multiply the
                # rhs. Doing * 2 - 1 ensures that when the event is crossed,
                # the sigmoid is zero. Hence the rhs is zero and the solution
                # stays constant for the rest of the simulation period
                # We can then cut off the part after the event was crossed
                event_sigmoid = (
                    pybamm.sigmoid(0, init_sign * event.expression, k) * 2 - 1
                )
                event_casadi = process(
                    event_sigmoid,
                    f"event_{n}",
                    vars_for_processing,
                    use_jacobian=False,
                )[0]
                # use the actual casadi object as this will go into the rhs
                casadi_switch_events.append(event_casadi)
            else:
                # use the function call
                event_call = process(
                    event.expression,
                    f"event_{n}",
                    vars_for_processing,
                    use_jacobian=False,
                )[0]
                if event.event_type == pybamm.EventType.TERMINATION:
                    terminate_events.append(event_call)
                elif event.event_type == pybamm.EventType.INTERPOLANT_EXTRAPOLATION:
                    interpolant_extrapolation_events.append(event_call)

        return (
            casadi_switch_events,
            terminate_events,
            interpolant_extrapolation_events,
            t_discon,
            discontinuity_events,
        )

    def _set_consistent_initialization(self, model, time, inputs_dict):
        """
        Set initialized states for the model. This is skipped if the solver is an
        algebraic solver (since this would make the algebraic solver redundant), and if
        the model doesn't have any algebraic equations (since there are no initial
        conditions to be calculated in this case).

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model for which to calculate initial conditions.
        time : numeric type
            The time at which to calculate the initial conditions
        inputs_dict : dict
            Any input parameters to pass to the model when solving
        update_rhs : bool
            Whether to update the rhs. True for 'solve', False for 'step'.

        """

        if self._algebraic_solver or model.len_alg == 0:
            # Don't update model.y0
            return

        # Calculate consistent states for the algebraic equations
        model.y0 = self.calculate_consistent_state(model, time, inputs_dict)

    def calculate_consistent_state(self, model, time=0, inputs=None):
        """
        Calculate consistent state for the algebraic equations through
        root-finding. model.y0 is used as the initial guess for rootfinding

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model for which to calculate initial conditions.
        time : float
            The time at which to calculate the initial conditions
        inputs: dict, optional
            Any input parameters to pass to the model when solving

        Returns
        -------
        y0_consistent : array-like, same shape as y0_guess
            Initial conditions that are consistent with the algebraic equations (roots
            of the algebraic equations). If self.root_method == None then returns
            model.y0.
        """
        pybamm.logger.debug("Start calculating consistent states")
        if self.root_method is None:
            return model.y0
        try:
            root_sol = self.root_method._integrate(model, np.array([time]), inputs)
        except pybamm.SolverError as e:
            raise pybamm.SolverError(
                f"Could not find consistent states: {e.args[0]}"
            ) from e
        pybamm.logger.debug("Found consistent states")

        self.check_extrapolation(root_sol, model.events)
        y0 = root_sol.all_ys[0]
        return y0

    def _solve_process_calculate_sensitivities_arg(
        inputs, model, calculate_sensitivities
    ):
        # get a list-only version of calculate_sensitivities
        if isinstance(calculate_sensitivities, bool):
            if calculate_sensitivities:
                calculate_sensitivities_list = [p for p in inputs.keys()]
            else:
                calculate_sensitivities_list = []
        else:
            calculate_sensitivities_list = calculate_sensitivities

        calculate_sensitivities_list.sort()
        if not hasattr(model, "calculate_sensitivities"):
            model.calculate_sensitivities = []

        # Check that calculate_sensitivites have not been updated
        sensitivities_have_changed = (
            calculate_sensitivities_list != model.calculate_sensitivities
        )

        # save sensitivity parameters so we can identify them later on
        # (FYI: this is used in the Solution class)
        model.calculate_sensitivities = calculate_sensitivities_list

        return calculate_sensitivities_list, sensitivities_have_changed

    def solve(
        self,
        model,
        t_eval=None,
        inputs=None,
        nproc=None,
        calculate_sensitivities=False,
        t_interp=None,
        initial_conditions=None,
    ):
        """
        Execute the solver setup and calculate the solution of the model at
        specified times.

        Parameters
        ----------
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate. Must have attributes rhs and
            initial_conditions. All calls to solve must pass in the same model or
            an error is raised
        t_eval : None, list or ndarray, optional
            The times (in seconds) at which to compute the solution. Defaults to None.
        inputs : dict or list, optional
            A dictionary or list of dictionaries describing any input parameters to
            pass to the model when solving
        nproc : int, optional
            Number of processes to use when solving for more than one set of input
            parameters. Defaults to value returned by "os.cpu_count()".
        calculate_sensitivities : list of str or bool, optional
            Whether the solver calculates sensitivities of all input parameters. Defaults to False.
            If only a subset of sensitivities are required, can also pass a
            list of input parameter names.  **Limitations**: sensitivities are not calculated up to numerical tolerances
            so are not guarenteed to be within the tolerances set by the solver, please raise an issue if you
            require this functionality. Also, when using this feature with `pybamm.Experiment`, the sensitivities
            do not take into account the movement of step-transitions wrt input parameters, so do not use this feature
            if the timings of your experimental protocol change rapidly with respect to your input parameters.
        t_interp : None, list or ndarray, optional
            The times (in seconds) at which to interpolate the solution. Defaults to None.
            Only valid for solvers that support intra-solve interpolation (`IDAKLUSolver`).
        initial_conditions : dict, numpy.ndarray, or list, optional
            Override the model’s default `y0`.  Can be:

            - a dict mapping variable names → values
            - a 1D array of length `n_states`
            - a list of such overrides (one per parallel solve)

            Only valid for IDAKLU solver.
        Returns
        -------
        :class:`pybamm.Solution` or list of :class:`pybamm.Solution` objects.
             If type of `inputs` is `list`, return a list of corresponding
             :class:`pybamm.Solution` objects.

        Raises
        ------
        :class:`pybamm.ModelError`
            If an empty model is passed (`model.rhs = {}` and `model.algebraic={}` and
            `model.variables = {}`)
        :class:`RuntimeError`
            If multiple calls to `solve` pass in different models

        """
        pybamm.logger.info(f"Start solving {model.name} with {self.name}")

        # Make sure model isn't empty
        self._check_empty_model(model)

        # t_eval can only be None if the solver is an algebraic solver. In that case
        # set it to 0
        if t_eval is None:
            if self._algebraic_solver is False:
                raise ValueError("t_eval cannot be None")
            t_eval = np.array([0])

        # If t_eval is provided as [t0, tf] return the solution at 100 points
        elif isinstance(t_eval, list):
            if len(t_eval) == 1 and self._algebraic_solver is True:
                t_eval = np.array(t_eval)
            elif len(t_eval) != 2:
                raise pybamm.SolverError(
                    "'t_eval' can be provided as an array of times at which to "
                    "return the solution, or as a list [t0, tf] where t0 is the "
                    "initial time and tf is the final time, but has been provided "
                    f"as a list of length {len(t_eval)}."
                )
            elif not self.supports_interp:
                t_eval = np.linspace(t_eval[0], t_eval[-1], 100)

        # Make sure t_eval is monotonic
        if (np.diff(t_eval) < 0).any():
            raise pybamm.SolverError("t_eval must increase monotonically")

        t_interp = self.process_t_interp(t_interp)

        # Set up inputs
        #
        # Argument "inputs" can be either a list of input dicts or
        # a single dict. The remaining of this function is only working
        # with variable "input_list", which is a list of dictionaries.
        # If "inputs" is a single dict, "inputs_list" is a list of only one dict.
        inputs_list = inputs if isinstance(inputs, list) else [inputs]
        model_inputs_list = [
            self._set_up_model_inputs(model, inputs) for inputs in inputs_list
        ]

        calculate_sensitivities_list, sensitivities_have_changed = (
            BaseSolver._solve_process_calculate_sensitivities_arg(
                model_inputs_list[0], model, calculate_sensitivities
            )
        )

        # (Re-)calculate consistent initialization
        # Assuming initial conditions do not depend on input parameters
        # when len(inputs_list) > 1, only `model_inputs_list[0]`
        # is passed to `_set_consistent_initialization`.
        # See https://github.com/pybamm-team/PyBaMM/pull/1261
        if len(model_inputs_list) > 1:
            all_inputs_names = set(
                itertools.chain.from_iterable(
                    [model_inputs.keys() for model_inputs in model_inputs_list]
                )
            )
            if all_inputs_names:
                initial_conditions_node_names = set(
                    [
                        it.name
                        for it in model.concatenated_initial_conditions.pre_order()
                    ]
                )
                if all_inputs_names.issubset(initial_conditions_node_names):
                    raise pybamm.SolverError(
                        "Input parameters cannot appear in expression "
                        "for initial conditions."
                    )

        # if any setup configuration has changed, we need to re-set up
        if sensitivities_have_changed:
            self._model_set_up.pop(model, None)
            # CasadiSolver caches its integrators using model, so delete this too
            if isinstance(self, pybamm.CasadiSolver):
                self.integrators.pop(model, None)

        # save sensitivity parameters so we can identify them later on
        # (FYI: this is used in the Solution class)
        model.calculate_sensitivities = calculate_sensitivities_list

        # Set up (if not done already)
        timer = pybamm.Timer()
        # Set the initial conditions
        if model not in self._model_set_up:
            if len(self._model_set_up) > 0:
                existing_model = next(iter(self._model_set_up))
                raise RuntimeError(
                    "This solver has already been initialised for model "
                    f'"{existing_model.name}". Please create a separate '
                    "solver for this model"
                )
            # It is assumed that when len(inputs_list) > 1, model set
            # up (initial condition, time-scale and length-scale) does
            # not depend on input parameters. Therefore, only `model_inputs[0]`
            # is passed to `set_up`.
            # See https://github.com/pybamm-team/PyBaMM/pull/1261
            self.set_up(model, model_inputs_list[0], t_eval)
            self._model_set_up.update(
                {model: {"initial conditions": model.concatenated_initial_conditions}}
            )
        elif (
            self._model_set_up[model]["initial conditions"]
            != model.concatenated_initial_conditions
        ):
            if self._algebraic_solver:
                # For an algebraic solver, we don't need to set up the initial
                # conditions function and we can just evaluate
                # model.concatenated_initial_conditions
                model.y0 = model.concatenated_initial_conditions.evaluate()
            else:
                # If the new initial conditions are different
                # and cannot be evaluated directly, set up again
                self.set_up(model, model_inputs_list[0], t_eval, ics_only=True)
            self._model_set_up[model]["initial conditions"] = (
                model.concatenated_initial_conditions
            )
        else:
            # Set the standard initial conditions
            self._set_initial_conditions(model, t_eval[0], model_inputs_list[0])

        # Solve for the consistent initialization
        self._set_consistent_initialization(model, t_eval[0], model_inputs_list[0])

        set_up_time = timer.time()
        timer.reset()

        # Check initial conditions don't violate events
        self._check_events_with_initialization(t_eval, model, model_inputs_list[0])

        # Process discontinuities
        (
            start_indices,
            end_indices,
            t_eval,
        ) = self._get_discontinuity_start_end_indices(model, inputs, t_eval)

        # Integrate separately over each time segment and accumulate into the solution
        # object, restarting the solver at each discontinuity (and recalculating a
        # consistent state afterwards if a DAE)
        old_y0 = model.y0
        solutions = None
        for start_index, end_index in zip(start_indices, end_indices, strict=False):
            pybamm.logger.verbose(
                f"Calling solver for {t_eval[start_index]} < t < {t_eval[end_index - 1]}"
            )
            if self.supports_parallel_solve:
                # Jax and IDAKLU solver can accept a list of inputs
                new_solutions = self._integrate(
                    model,
                    t_eval[start_index:end_index],
                    model_inputs_list,
                    t_interp,
                    initial_conditions,
                )
            else:
                ninputs = len(model_inputs_list)
                if ninputs == 1:
                    new_solution = self._integrate(
                        model,
                        t_eval[start_index:end_index],
                        model_inputs_list[0],
                        t_interp=t_interp,
                    )
                    new_solutions = [new_solution]
                else:
                    with mp.get_context(self._mp_context).Pool(processes=nproc) as p:
                        new_solutions = p.starmap(
                            self._integrate,
                            zip(
                                [model] * ninputs,
                                [t_eval[start_index:end_index]] * ninputs,
                                model_inputs_list,
                                [t_interp] * ninputs,
                                strict=True,
                            ),
                        )
                        p.close()
                        p.join()
            # Setting the solve time for each segment.
            # pybamm.Solution.__add__ assumes attribute solve_time.
            solve_time = timer.time()
            for sol in new_solutions:
                sol.solve_time = solve_time
            if start_index == start_indices[0]:
                solutions = [sol for sol in new_solutions]
            else:
                for i, new_solution in enumerate(new_solutions):
                    solutions[i] = solutions[i] + new_solution

            if solutions[0].termination != "final time":
                break

            if end_index != len(t_eval):
                # setup for next integration subsection
                last_state = solutions[0].y[:, -1]
                # update y0 (for DAE solvers, this updates the initial guess for the
                # rootfinder)
                model.y0 = last_state
                if len(model.algebraic) > 0:
                    model.y0 = self.calculate_consistent_state(
                        model, t_eval[end_index], model_inputs_list[0]
                    )
        solve_time = timer.time()

        for i, solution in enumerate(solutions):
            # Check if extrapolation occurred
            self.check_extrapolation(solution, model.events)
            # Identify the event that caused termination and update the solution to
            # include the event time and state
            solutions[i], termination = self.get_termination_reason(
                solution, model.events
            )
            # Assign times
            solutions[i].set_up_time = set_up_time
            # all solutions get the same solve time, but their integration time
            # will be different (see https://github.com/pybamm-team/PyBaMM/pull/1261)
            solutions[i].solve_time = solve_time

        # Restore old y0
        model.y0 = old_y0

        # Report times
        if len(solutions) == 1:
            pybamm.logger.info(f"Finish solving {model.name} ({termination})")
            pybamm.logger.info(
                f"Set-up time: {solutions[0].set_up_time}, Solve time: {solutions[0].solve_time} (of which integration time: {solutions[0].integration_time}), "
                f"Total time: {solutions[0].total_time}"
            )
        else:
            pybamm.logger.info(f"Finish solving {model.name} for all inputs")
            pybamm.logger.info(
                f"Set-up time: {solutions[0].set_up_time}, Solve time: {solutions[0].solve_time}, Total time: {solutions[0].total_time}"
            )

        # Raise error if solutions[0] only contains one timestep (except for algebraic
        # solvers, where we may only expect one time in the solution)
        if (
            self._algebraic_solver is False
            and len(solutions[0].all_ts) == 1
            and len(solutions[0].all_ts[0]) == 1
        ):
            raise pybamm.SolverError(
                "Solution time vector has length 1. "
                "Check whether simulation terminated too early."
            )

        # Return solution(s)
        if len(solutions) == 1:
            return solutions[0]
        else:
            return solutions

    @staticmethod
    def filter_discontinuities(t_discon: list, t_eval: list) -> np.ndarray:
        """
        Filter the discontinuities to only include the unique and sorted
        values within the t_eval range (non-exclusive of end points).

        Parameters
        ----------
        t_discon : list
            The list of all possible discontinuity times.
        t_eval : list
            The integration time points.

        Returns
        -------
        np.ndarray
            The filtered list of discontinuities within the range of t_eval.
        """
        t_discon_unique = np.unique(t_discon)

        # Find the indices within t_eval (non-exclusive of end points)
        idx_start = np.searchsorted(t_discon_unique, t_eval[0], side="right")
        idx_end = np.searchsorted(t_discon_unique, t_eval[-1], side="left")
        return t_discon_unique[idx_start:idx_end]

    def _get_discontinuity_start_end_indices(self, model, inputs, t_eval):
        if self.supports_t_eval_discontinuities:
            t_discon_constant = self.filter_discontinuities(
                model.t_discon_constant, t_eval
            )
            t_eval = np.union1d(t_eval, t_discon_constant)

        if not model.discontinuity_events_eval:
            pybamm.logger.verbose("No discontinuity events found")
            return [0], [len(t_eval)], t_eval

        # Calculate all possible discontinuities
        _t_discon_full = [
            # Assuming that discontinuities do not depend on
            # input parameters when len(input_list) > 1, only
            # `inputs` is passed to `evaluate`.
            # See https://github.com/pybamm-team/PyBaMM/pull/1261
            event.expression.evaluate(inputs=inputs)
            for event in model.discontinuity_events_eval
        ]
        t_discon = self.filter_discontinuities(_t_discon_full, t_eval)

        pybamm.logger.verbose(f"Discontinuity events found at t = {t_discon}")
        if isinstance(inputs, list):
            raise pybamm.SolverError(
                "Cannot solve for a list of input parameters sets with discontinuities"
            )

        # insert time points around discontinuities in t_eval
        # keep track of subsections to integrate by storing start and end indices
        start_indices = [0]
        end_indices = []
        eps = sys.float_info.epsilon
        for dtime in t_discon:
            dindex = np.searchsorted(t_eval, dtime, side="left")
            end_indices.append(dindex + 1)
            start_indices.append(dindex + 1)
            if dtime * (1 - eps) < t_eval[dindex] < dtime * (1 + eps):
                t_eval[dindex] *= 1 + eps
                t_eval = np.insert(t_eval, dindex, dtime * (1 - eps))
            else:
                t_eval = np.insert(
                    t_eval, dindex, [dtime * (1 - eps), dtime * (1 + eps)]
                )
        end_indices.append(len(t_eval))

        return start_indices, end_indices, t_eval

    @staticmethod
    def _check_events_with_initialization(t_eval, model, inputs_dict):
        num_terminate_events = len(model.terminate_events_eval)
        if num_terminate_events == 0:
            return

        if model.convert_to_format == "casadi":
            inputs = casadi.vertcat(*[x for x in inputs_dict.values()])

        events_eval = np.empty(num_terminate_events)
        for idx, event in enumerate(model.terminate_events_eval):
            if model.convert_to_format == "casadi":
                event_eval = event(t_eval[0], model.y0, inputs)
            elif model.convert_to_format in ["python", "jax"]:
                event_eval = event(t=t_eval[0], y=model.y0, inputs=inputs_dict)
            events_eval[idx] = event_eval

        if events_eval.min() <= 0:
            # find the events that were triggered by initial conditions
            termination_events = [
                x for x in model.events if x.event_type == pybamm.EventType.TERMINATION
            ]
            idxs = np.where(events_eval <= 0)[0]
            event_names = [termination_events[idx].name for idx in idxs]
            raise pybamm.SolverError(
                f"Events {event_names} are non-positive at initial conditions"
            )

    def _set_sens_initial_conditions_from(
        self, solution: pybamm.Solution, model: pybamm.BaseModel
    ) -> tuple:
        """
        A restricted version of BaseModel.set_initial_conditions_from that only extracts the
        sensitivities from a solution object, and only for a model that has been discretised.
        This is used when setting the initial conditions for a sensitivity model.

        Parameters
        ----------
        solution : :class:`pybamm.Solution`
            The solution to use to initialize the model

        model: :class:`pybamm.BaseModel`
            The model whose sensitivities to set

        Returns
        -------

        initial_conditions : tuple of ndarray
            The initial conditions for the sensitivities, each element of the tuple
            corresponds to an input parameter
        """

        ninputs = len(model.calculate_sensitivities)
        initial_conditions = tuple([] for _ in range(ninputs))
        solution = solution.last_state
        for var in model.initial_conditions:
            final_state = solution[var.name]
            final_state = final_state.sensitivities
            final_state_eval = tuple(
                final_state[key] for key in model.calculate_sensitivities
            )

            scale, reference = var.scale.value, var.reference.value
            for i in range(ninputs):
                scaled_final_state_eval = (final_state_eval[i] - reference) / scale
                initial_conditions[i].append(scaled_final_state_eval)

        # Also update the concatenated initial conditions if the model is already
        # discretised
        # Unpack slices for sorting
        y_slices = {var: slce for var, slce in model.y_slices.items()}
        slices = [y_slices[symbol][0] for symbol in model.initial_conditions.keys()]

        # sort equations according to slices
        concatenated_initial_conditions = [
            casadi.vertcat(*[eq for _, eq in sorted(zip(slices, init, strict=True))])
            for init in initial_conditions
        ]
        return concatenated_initial_conditions

    def process_t_interp(self, t_interp):
        # set a variable for this
        no_interp = (not self.supports_interp) and (
            t_interp is not None and len(t_interp) != 0
        )
        if no_interp:
            warnings.warn(
                f"Explicit interpolation times not implemented for {self.name}",
                pybamm.SolverWarning,
                stacklevel=2,
            )

        if no_interp or t_interp is None:
            t_interp = np.empty(0)

        return t_interp

    def step(
        self,
        old_solution,
        model,
        dt,
        t_eval=None,
        npts=None,
        inputs=None,
        save=True,
        calculate_sensitivities=False,
        t_interp=None,
    ):
        """
        Step the solution of the model forward by a given time increment. The
        first time this method is called it executes the necessary setup by
        calling `self.set_up(model)`.

        Parameters
        ----------
        old_solution : :class:`pybamm.Solution` or None
            The previous solution to be added to. If `None`, a new solution is created.
        model : :class:`pybamm.BaseModel`
            The model whose solution to calculate. Must have attributes rhs and
            initial_conditions
        dt : numeric type
            The timestep (in seconds) over which to step the solution
        t_eval : list or numpy.ndarray, optional
            An array of times at which to stop the simulation and return the solution during the step
            (Note: t_eval is the time measured from the start of the step, so should start at 0 and end at dt).
            By default, the solution is returned at t0 and t0 + dt.
        npts : deprecated
        inputs : dict, optional
            Any input parameters to pass to the model when solving
        save : bool, optional
            Save solution with all previous timesteps. Defaults to True.
        calculate_sensitivities : list of str or bool, optional
            Whether the solver calculates sensitivities of all input parameters. Defaults to False.
            If only a subset of sensitivities are required, can also pass a
            list of input parameter names. **Limitations**: sensitivities are not calculated up to numerical tolerances
            so are not guarenteed to be within the tolerances set by the solver, please raise an issue if you
            require this functionality. Also, when using this feature with `pybamm.Experiment`, the sensitivities
            do not take into account the movement of step-transitions wrt input parameters, so do not use this feature
            if the timings of your experimental protocol change rapidly with respect to your input parameters.
        t_interp : None, list or ndarray, optional
            The times (in seconds) at which to interpolate the solution. Defaults to None.
            Only valid for solvers that support intra-solve interpolation (`IDAKLUSolver`).
        Raises
        ------
        :class:`pybamm.ModelError`
            If an empty model is passed (`model.rhs = {}` and `model.algebraic = {}` and
            `model.variables = {}`)

        """
        if old_solution is None:
            old_solution = pybamm.EmptySolution()

        if not (
            isinstance(old_solution, pybamm.EmptySolution)
            or old_solution.termination == "final time"
            or "[experiment]" in old_solution.termination
        ):
            # Return same solution as an event has already been triggered
            # With hack to allow stepping past experiment current / voltage cut-off
            return old_solution

        # Make sure model isn't empty
        self._check_empty_model(model)

        # Make sure dt is greater than zero
        if dt <= 0:
            raise pybamm.SolverError("Step time must be >0")

        # Raise deprecation warning for npts and convert it to t_eval
        if npts is not None:
            warnings.warn(
                "The 'npts' parameter is deprecated, use 't_eval' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            t_eval = np.linspace(0, dt, npts)
        elif t_eval is None:
            t_eval = np.array([0, dt])
        elif t_eval[0] != 0:
            raise pybamm.SolverError(
                f"The first `t_eval` value ({t_eval[0]}) must be 0."
                "Please correct your `t_eval` array."
            )
        elif t_eval[-1] != dt:
            raise pybamm.SolverError(
                f"The final `t_eval` value ({t_eval[-1]}) must be equal "
                f"to the step time `dt` ({dt}). Please correct your `t_eval` array."
            )
        else:
            pass

        t_interp = self.process_t_interp(t_interp)

        t_start = old_solution.t[-1]
        t_eval = t_start + t_eval
        t_interp = t_start + t_interp
        t_end = t_start + dt

        if t_start == 0:
            t_start_shifted = t_start
        else:
            # find the next largest floating point value for t_start
            # to avoid repeated times in the solution
            # from having the same time at the end of the previous step and
            # the start of the next step
            t_start_shifted = np.nextafter(t_start, np.inf)
            t_eval[0] = t_start_shifted
            if t_interp.size > 0 and t_interp[0] == t_start:
                t_interp[0] = t_start_shifted

        # Set timer
        timer = pybamm.Timer()

        # Set up inputs
        model_inputs = self._set_up_model_inputs(model, inputs)

        # process calculate_sensitivities argument
        calculate_sensitivities_list, sensitivities_have_changed = (
            BaseSolver._solve_process_calculate_sensitivities_arg(
                model_inputs, model, calculate_sensitivities
            )
        )

        first_step_this_model = model not in self._model_set_up
        if first_step_this_model or sensitivities_have_changed:
            if len(self._model_set_up) > 0:
                existing_model = next(iter(self._model_set_up))
                raise RuntimeError(
                    "This solver has already been initialised for model "
                    f'"{existing_model.name}". Please create a separate '
                    "solver for this model"
                )
            self.set_up(model, model_inputs)
            self._model_set_up.update(
                {model: {"initial conditions": model.concatenated_initial_conditions}}
            )

        if (
            isinstance(old_solution, pybamm.EmptySolution)
            and old_solution.termination is None
        ):
            pybamm.logger.verbose(f"Start stepping {model.name} with {self.name}")

        using_sensitivities = len(model.calculate_sensitivities) > 0

        if isinstance(old_solution, pybamm.EmptySolution):
            if not first_step_this_model:
                # reset y0 to original initial conditions
                self.set_up(model, model_inputs, ics_only=True)
        elif old_solution.all_models[-1] == model:
            last_state = old_solution.last_state
            model.y0 = last_state.all_ys[0]
            if using_sensitivities:
                full_sens = last_state._all_sensitivities["all"][0]
                model.y0S = tuple(full_sens[:, i] for i in range(full_sens.shape[1]))

        else:
            _, concatenated_initial_conditions = model.set_initial_conditions_from(
                old_solution,
                inputs=model_inputs,
                return_type="ics",
            )
            model.y0 = concatenated_initial_conditions.evaluate(0, inputs=model_inputs)
            if using_sensitivities:
                model.y0S = self._set_sens_initial_conditions_from(old_solution, model)

        set_up_time = timer.time()

        # (Re-)calculate consistent initialization
        self._set_consistent_initialization(model, t_start_shifted, model_inputs)

        # Check consistent initialization doesn't violate events
        self._check_events_with_initialization(t_eval, model, model_inputs)

        # Step
        pybamm.logger.verbose(f"Stepping for {t_start_shifted:.0f} < t < {t_end:.0f}")
        timer.reset()

        # API for _integrate is different for JaxSolver and IDAKLUSolver
        if self.supports_parallel_solve:
            solutions = self._integrate(model, t_eval, [model_inputs], t_interp)
            solution = solutions[0]
        else:
            solution = self._integrate(model, t_eval, model_inputs, t_interp)
        solution.solve_time = timer.time()

        # Check if extrapolation occurred
        self.check_extrapolation(solution, model.events)

        # Identify the event that caused termination and update the solution to
        # include the event time and state
        solution, termination = self.get_termination_reason(solution, model.events)

        # Assign setup time
        solution.set_up_time = set_up_time

        # Report times
        pybamm.logger.verbose(f"Finish stepping {model.name} ({termination})")
        pybamm.logger.verbose(
            f"Set-up time: {solution.set_up_time}, Step time: {solution.solve_time} (of which integration time: {solution.integration_time}), "
            f"Total time: {solution.total_time}"
        )

        # Return solution
        if save is False:
            return solution
        else:
            return old_solution + solution

    @staticmethod
    def get_termination_reason(solution, events):
        """
        Identify the cause for termination. In particular, if the solver terminated
        due to an event, (try to) pinpoint which event was responsible. If an event
        occurs the event time and state are added to the solution object.
        Note that the current approach (evaluating all the events and then finding which
        one is smallest at the final timestep) is pretty crude, but is the easiest one
        that works for all the different solvers.

        Parameters
        ----------
        solution : :class:`pybamm.Solution`
            The solution object
        events : dict
            Dictionary of events
        """
        termination_events = [
            x for x in events if x.event_type == pybamm.EventType.TERMINATION
        ]

        match solution.termination:
            case "final time":
                return (
                    solution,
                    "the solver successfully reached the end of the integration interval",
                )
            case "failure":
                return (
                    solution,
                    "the solver failed to simulate",
                )

        # solution.termination == "event":
        pybamm.logger.debug("Start post-processing events")
        if solution.closest_event_idx is not None:
            solution.termination = (
                f"event: {termination_events[solution.closest_event_idx].name}"
            )
        else:
            # Get final event value
            final_event_values = {}

            for event in termination_events:
                final_event_values[event.name] = event.expression.evaluate(
                    solution.t_event,
                    solution.y_event,
                    inputs=solution.all_inputs[-1],
                )
            termination_event = min(final_event_values, key=final_event_values.get)

            # Check that it's actually an event
            if final_event_values[termination_event] > 0.1:  # pragma: no cover
                # Hard to test this
                raise pybamm.SolverError(
                    "Could not determine which event was triggered "
                    "(possibly due to NaNs)"
                )
            # Add the event to the solution object
            solution.termination = f"event: {termination_event}"
        # Update t, y and inputs to include event time and state
        # Note: if the final entry of t is equal to the event time we skip
        # this (having duplicate entries causes an error later in ProcessedVariable)
        if solution.t_event != solution.all_ts[-1][-1]:
            event_sol = pybamm.Solution(
                solution.t_event,
                solution.y_event,
                solution.all_models[-1],
                solution.all_inputs[-1],
                solution.t_event,
                solution.y_event,
                solution.termination,
                variables_returned=solution.variables_returned,
            )
            event_sol.solve_time = 0
            event_sol.integration_time = 0
            solution = solution + event_sol

        pybamm.logger.debug("Finish post-processing events")
        return solution, solution.termination

    def check_extrapolation(self, solution, events):
        """
        Check if extrapolation occurred for any of the interpolants. Note that with the
        current approach (evaluating all the events at the solution times) some
        extrapolations might not be found if they only occurred for a small period of
        time.

        Parameters
        ----------
        solution : :class:`pybamm.Solution`
            The solution object
        events : dict
            Dictionary of events
        """
        if self.on_extrapolation == "ignore":
            return

        extrap_events = []

        # Add the event dictionary to the solution object
        solution.extrap_events = extrap_events

        # first pass: check if any events are extrapolation events
        if all(
            event.event_type != pybamm.EventType.INTERPOLANT_EXTRAPOLATION
            for event in events
        ):
            # no extrapolation events to check
            return

        # second pass: check if the extrapolation events are within the tolerance
        last_state = solution.last_state
        if solution.t_event:
            t = solution.t_event[0]
            y = solution.y_event[:, 0]
        else:
            t = last_state.all_ts[0][0]
            y = last_state.all_ys[0][:, 0]
        inputs = last_state.all_inputs[0]

        if isinstance(y, casadi.DM):
            y = y.full()

        for event in events:
            if (
                event.event_type == pybamm.EventType.INTERPOLANT_EXTRAPOLATION
                and event.expression.evaluate(t, y, inputs=inputs) < self.extrap_tol
            ):
                extrap_events.append(event.name)

        if len(extrap_events) == 0:
            # no extrapolation events are within the tolerance
            return

        if self.on_extrapolation == "error":
            raise pybamm.SolverError(
                "Solver failed because the following "
                f"interpolation bounds were exceeded: {extrap_events}. "
                "You may need to provide additional interpolation points "
                "outside these bounds."
            )
        elif self.on_extrapolation == "warn":
            name = solution.all_models[-1].name
            warnings.warn(
                f"While solving {name} extrapolation occurred for {extrap_events}",
                pybamm.SolverWarning,
                stacklevel=2,
            )

    def _check_empty_model(self, model):
        # Make sure model isn't empty
        if (
            (len(model.rhs) == 0 and len(model.algebraic) == 0)
            and model.concatenated_rhs is None
            and model.concatenated_algebraic is None
            and not isinstance(self, pybamm.DummySolver)
        ):
            raise pybamm.ModelError(
                "Cannot simulate an empty model, use `pybamm.DummySolver` instead"
            )

    def get_platform_context(self, system_type: str):
        # Set context for parallel processing depending on the platform
        if system_type.lower() in ["linux"]:
            return "fork"
        return "spawn"

    @staticmethod
    def _set_up_model_inputs(model, inputs):
        """Set up input parameters"""
        if inputs is None:
            inputs = {}
        else:
            inputs = ParameterValues.check_parameter_values(inputs)

        # Go through all input parameters that can be found in the model
        # Only keep the ones that are actually used in the model
        # If any of them are *not* provided by "inputs", raise an error
        inputs_in_model = {}
        for input_param in model.input_parameters:
            name = input_param.name
            if name not in inputs:
                raise pybamm.SolverError(f"No value provided for input '{name}'")
            inputs_in_model[name] = inputs[name]

        inputs = inputs_in_model

        ordered_inputs_names = list(inputs.keys())
        ordered_inputs_names.sort()
        ordered_inputs = {name: inputs[name] for name in ordered_inputs_names}

        return ordered_inputs


def process(
    symbol, name, vars_for_processing, use_jacobian=None, return_jacp_stacked=None
):
    """
    Parameters
    ----------
    symbol: :class:`pybamm.Symbol`
        expression tree to convert
    name: str
        function evaluators created will have this base name
    use_jacobian: bool, optional
        whether to return Jacobian functions
    return_jacp_stacked: bool, optional
        returns Jacobian function wrt stacked parameters instead of jacp

    Returns
    -------
    func: :class:`pybamm.EvaluatorPython` or
            :class:`pybamm.EvaluatorJax` or
            :class:`casadi.Function`
        evaluator for the function $f(y, t, p)$ given by `symbol`

    jac: :class:`pybamm.EvaluatorPython` or
            :class:`pybamm.EvaluatorJaxJacobian` or
            :class:`casadi.Function`
        evaluator for the Jacobian $\\frac{\\partial f}{\\partial y}$
        of the function given by `symbol`

    jacp: :class:`pybamm.EvaluatorPython` or
            :class:`pybamm.EvaluatorJaxSensitivities` or
            :class:`casadi.Function`
        evaluator for the parameter sensitivities
        $\frac{\\partial f}{\\partial p}$
        of the function given by `symbol`

    jac_action: :class:`pybamm.EvaluatorPython` or
            :class:`pybamm.EvaluatorJax` or
            :class:`casadi.Function`
        evaluator for product of the Jacobian with a vector $v$,
        i.e. $\\frac{\\partial f}{\\partial y} * v$
    """

    def report(string):
        # don't log event conversion
        if "event" not in string:
            pybamm.logger.verbose(string)

    model = vars_for_processing["model"]

    if use_jacobian is None:
        use_jacobian = model.use_jacobian

    if model.convert_to_format == "jax":
        report(f"Converting {name} to jax")
        func = pybamm.EvaluatorJax(symbol)
        jacp = None
        if model.calculate_sensitivities:
            report(
                f"Calculating sensitivities for {name} with respect "
                f"to parameters {model.calculate_sensitivities} using jax"
            )
            jacp = func.get_sensitivities()
        if use_jacobian:
            report(f"Calculating jacobian for {name} using jax")
            jac = func.get_jacobian()
            jac_action = func.get_jacobian_action()
        else:
            jac = None
            jac_action = None

    elif model.convert_to_format != "casadi":
        y = vars_for_processing["y"]
        jacobian = vars_for_processing["jacobian"]

        if model.calculate_sensitivities:
            raise pybamm.SolverError(  # pragma: no cover
                "Sensitivies are no longer supported for the python "
                "evaluator. Please use `convert_to_format = 'casadi'`, or `jax` "
                "to calculate sensitivities."
            )

        else:
            jacp = None

        if use_jacobian:
            report(f"Calculating jacobian for {name}")
            jac = jacobian.jac(symbol, y)
            report(f"Converting jacobian for {name} to python")
            jac = pybamm.EvaluatorPython(jac)
            # cannot do jacobian action efficiently for now
            jac_action = None
        else:
            jac = None
            jac_action = None

        report(f"Converting {name} to python")
        func = pybamm.EvaluatorPython(symbol)

    else:
        t_casadi = vars_for_processing["t_casadi"]
        y_casadi = vars_for_processing["y_casadi"]
        p_casadi = vars_for_processing["p_casadi"]
        p_casadi_stacked = vars_for_processing["p_casadi_stacked"]

        # Process with CasADi
        report(f"Converting {name} to CasADi")
        casadi_expression = symbol.to_casadi(t_casadi, y_casadi, inputs=p_casadi)
        # Add sensitivity vectors to the rhs and algebraic equations
        jacp = None

        if model.calculate_sensitivities:
            report(
                f"Calculating sensitivities for {name} with respect "
                f"to parameters {model.calculate_sensitivities} using "
                "CasADi"
            )
            # Compute derivate wrt p-stacked (can be passed to solver to
            # compute sensitivities online)
            if return_jacp_stacked:
                jacp = casadi.Function(
                    f"d{name}_dp",
                    [t_casadi, y_casadi, p_casadi_stacked],
                    [casadi.jacobian(casadi_expression, p_casadi_stacked)],
                )
            else:
                # WARNING, jacp for convert_to_format=casadi does not return a dict
                # instead it returns multiple return values, one for each param
                # TODO: would it be faster to do the jacobian wrt pS_casadi_stacked?
                jacp = casadi.Function(
                    name + "_jacp",
                    [t_casadi, y_casadi, p_casadi_stacked],
                    [
                        casadi.densify(
                            casadi.jacobian(casadi_expression, p_casadi[pname])
                        )
                        for pname in model.calculate_sensitivities
                    ],
                )

        if use_jacobian:
            report(f"Calculating jacobian for {name} using CasADi")
            jac_casadi = casadi.jacobian(casadi_expression, y_casadi)
            jac = casadi.Function(
                name + "_jac",
                [t_casadi, y_casadi, p_casadi_stacked],
                [jac_casadi],
            )

            v = casadi.MX.sym(
                "v",
                model.len_rhs_and_alg,
            )
            jac_action_casadi = casadi.densify(
                casadi.jtimes(casadi_expression, y_casadi, v)
            )
            jac_action = casadi.Function(
                name + "_jac_action",
                [t_casadi, y_casadi, p_casadi_stacked, v],
                [jac_action_casadi],
            )
        else:
            jac = None
            jac_action = None

        func = casadi.Function(
            name, [t_casadi, y_casadi, p_casadi_stacked], [casadi_expression]
        )

    return func, jac, jacp, jac_action



#
# Class for particles with Fickian diffusion
#
import pybamm

from .base_particle import BaseParticle


class FickianDiffusion(BaseParticle):
    """
    Class for molar conservation in particles, employing Fick's law

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain of the model either 'Negative' or 'Positive'
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    x_average : bool
        Whether the particle concentration is averaged over the x-direction
    """

    def __init__(self, param, domain, options, phase="primary", x_average=False):
        super().__init__(param, domain, options, phase)
        self.x_average = x_average

    def get_fundamental_variables(self):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name
        Phase_prefactor = self.phase_param.phase_prefactor
        variables = {}
        if self.size_distribution is False:
            if self.x_average is False:
                c_s = pybamm.Variable(
                    f"{Domain} {phase_name}particle concentration [mol.m-3]",
                    f"{domain} {phase_name}particle",
                    auxiliary_domains={
                        "secondary": f"{domain} electrode",
                        "tertiary": "current collector",
                    },
                    bounds=(0, self.phase_param.c_max),
                    scale=self.phase_param.c_max,
                )
                c_s.print_name = f"c_s_{domain[0]}"
            else:
                c_s_xav = pybamm.Variable(
                    f"X-averaged {domain} {phase_name}particle concentration [mol.m-3]",
                    f"{domain} {phase_name}particle",
                    auxiliary_domains={"secondary": "current collector"},
                    bounds=(0, self.phase_param.c_max),
                    scale=self.phase_param.c_max,
                )
                c_s_xav.print_name = f"c_s_{domain[0]}_xav"
                c_s = pybamm.SecondaryBroadcast(c_s_xav, f"{domain} electrode")
        else:
            if self.x_average is False:
                c_s_distribution = pybamm.Variable(
                    f"{Domain} {phase_name}particle "
                    "concentration distribution [mol.m-3]",
                    domain=f"{domain} {phase_name}particle",
                    auxiliary_domains={
                        "secondary": f"{domain} {phase_name}particle size",
                        "tertiary": f"{domain} electrode",
                        "quaternary": "current collector",
                    },
                    bounds=(0, self.phase_param.c_max),
                    scale=self.phase_param.c_max,
                )
                R = pybamm.SpatialVariable(
                    f"R_{domain[0]}",
                    domain=[f"{domain} {phase_name}particle size"],
                    auxiliary_domains={
                        "secondary": f"{domain} electrode",
                        "tertiary": "current collector",
                    },
                    coord_sys="cartesian",
                )
                variables = self._get_distribution_variables(R)
                f_v_dist = variables[
                    f"{Phase_prefactor}{Domain} volume-weighted {phase_name}"
                    "particle-size distribution [m-1]"
                ]
            else:
                c_s_distribution = pybamm.Variable(
                    f"X-averaged {domain} {phase_name}particle "
                    "concentration distribution [mol.m-3]",
                    domain=f"{domain} {phase_name}particle",
                    auxiliary_domains={
                        "secondary": f"{domain} {phase_name}particle size",
                        "tertiary": "current collector",
                    },
                    bounds=(0, self.phase_param.c_max),
                    scale=self.phase_param.c_max,
                )
                R = pybamm.SpatialVariable(
                    f"R_{domain[0]}",
                    domain=[f"{domain} {phase_name}particle size"],
                    auxiliary_domains={"secondary": "current collector"},
                    coord_sys="cartesian",
                )
                variables = self._get_distribution_variables(R)
                f_v_dist = variables[
                    f"X-averaged {domain} volume-weighted {phase_name}"
                    "particle-size distribution [m-1]"
                ]

            # Standard concentration distribution variables (size-dependent)
            variables.update(
                self._get_standard_concentration_distribution_variables(
                    c_s_distribution
                )
            )
            # Standard size-averaged variables. Average concentrations using
            # the volume-weighted distribution since they are volume-based
            # quantities. Necessary for output variables "Total lithium in
            # negative electrode [mol]", etc, to be calculated correctly
            c_s = pybamm.Integral(f_v_dist * c_s_distribution, R)
            if self.x_average is True:
                c_s = pybamm.SecondaryBroadcast(c_s, [f"{domain} electrode"])

        # Standard concentration variables (size-independent)
        variables.update(self._get_standard_concentration_variables(c_s))

        return variables

    def get_coupled_variables(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        if self.size_distribution is False:
            if self.x_average is False:
                c_s = variables[
                    f"{Domain} {phase_name}particle concentration [mol.m-3]"
                ]
                T = pybamm.PrimaryBroadcast(
                    variables[f"{Domain} electrode temperature [K]"],
                    [f"{domain} {phase_name}particle"],
                )
                R_nondim = variables[f"{Domain} {phase_name}particle radius"]
                j = variables[
                    f"{Domain} electrode {phase_name}"
                    "interfacial current density [A.m-2]"
                ]
            else:
                c_s = variables[
                    f"X-averaged {domain} {phase_name}particle concentration [mol.m-3]"
                ]
                T = pybamm.PrimaryBroadcast(
                    variables[f"X-averaged {domain} electrode temperature [K]"],
                    [f"{domain} {phase_name}particle"],
                )
                R_nondim = 1
                j = variables[
                    f"X-averaged {domain} electrode {phase_name}"
                    "interfacial current density [A.m-2]"
                ]
            R_broad_nondim = R_nondim
        else:
            R_nondim = variables[f"{Domain} {phase_name}particle sizes"]
            R_broad_nondim = pybamm.PrimaryBroadcast(
                R_nondim, [f"{domain} {phase_name}particle"]
            )
            if self.x_average is False:
                c_s = variables[
                    f"{Domain} {phase_name}particle "
                    "concentration distribution [mol.m-3]"
                ]
                # broadcast T to "particle size" domain then again into "particle"
                T = pybamm.PrimaryBroadcast(
                    variables[f"{Domain} electrode temperature [K]"],
                    [f"{domain} {phase_name}particle size"],
                )
                T = pybamm.PrimaryBroadcast(T, [f"{domain} {phase_name}particle"])
                j = variables[
                    f"{Domain} electrode {phase_name}interfacial "
                    "current density distribution [A.m-2]"
                ]
            else:
                c_s = variables[
                    f"X-averaged {domain} {phase_name}particle "
                    "concentration distribution [mol.m-3]"
                ]
                # broadcast to "particle size" domain then again into "particle"
                T = pybamm.PrimaryBroadcast(
                    variables[f"X-averaged {domain} electrode temperature [K]"],
                    [f"{domain} {phase_name}particle size"],
                )
                T = pybamm.PrimaryBroadcast(T, [f"{domain} {phase_name}particle"])
                j = variables[
                    f"X-averaged {domain} electrode {phase_name}interfacial "
                    "current density distribution [A.m-2]"
                ]

        current = variables["Total current density [A.m-2]"]
        D_eff = self._get_effective_diffusivity(c_s, T, current)
        N_s = -D_eff * pybamm.grad(c_s)

        variables.update(
            {
                f"{Domain} {phase_name}particle rhs [mol.m-3.s-1]": -(
                    1 / (R_broad_nondim**2)
                )
                * pybamm.div(N_s),
                f"{Domain} {phase_name}particle bc [mol.m-4]": -j
                * R_nondim
                / self.param.F
                / pybamm.surf(D_eff),
            }
        )

        if self.x_average is True:
            if self.size_distribution is True:
                D_eff = pybamm.TertiaryBroadcast(D_eff, [f"{domain} electrode"])
                N_s = pybamm.TertiaryBroadcast(N_s, [f"{domain} electrode"])
            else:
                D_eff = pybamm.SecondaryBroadcast(D_eff, [f"{domain} electrode"])
                N_s = pybamm.SecondaryBroadcast(N_s, [f"{domain} electrode"])

        if self.size_distribution is True:
            # Size-dependent flux variables
            variables.update(
                self._get_standard_diffusivity_distribution_variables(D_eff)
            )
            variables.update(self._get_standard_flux_distribution_variables(N_s))
            # Size-averaged flux variables
            D_eff = pybamm.size_average(D_eff)
            R = pybamm.SpatialVariable("R", domains=D_eff.domains)
            f_a_dist = self.phase_param.f_a_dist(R)
            N_s = pybamm.Integral(f_a_dist * N_s, R)
        variables.update(self._get_standard_diffusivity_variables(D_eff))
        variables.update(self._get_standard_flux_variables(N_s))

        return variables

    def set_rhs(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        if self.size_distribution is False:
            if self.x_average is False:
                c_s = variables[
                    f"{Domain} {phase_name}particle concentration [mol.m-3]"
                ]
            else:
                c_s = variables[
                    f"X-averaged {domain} {phase_name}particle concentration [mol.m-3]"
                ]
        else:
            if self.x_average is False:
                c_s = variables[
                    f"{Domain} {phase_name}particle "
                    "concentration distribution [mol.m-3]"
                ]
            else:
                c_s = variables[
                    f"X-averaged {domain} {phase_name}particle "
                    "concentration distribution [mol.m-3]"
                ]
        self.rhs = {c_s: variables[f"{Domain} {phase_name}particle rhs [mol.m-3.s-1]"]}

    def set_boundary_conditions(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        if self.size_distribution is False:
            if self.x_average is False:
                c_s = variables[
                    f"{Domain} {phase_name}particle concentration [mol.m-3]"
                ]
            else:
                c_s = variables[
                    f"X-averaged {domain} {phase_name}particle concentration [mol.m-3]"
                ]
        else:
            if self.x_average is False:
                c_s = variables[
                    f"{Domain} {phase_name}particle "
                    "concentration distribution [mol.m-3]"
                ]
            else:
                c_s = variables[
                    f"X-averaged {domain} {phase_name}particle "
                    "concentration distribution [mol.m-3]"
                ]

        rbc = variables[f"{Domain} {phase_name}particle bc [mol.m-4]"]
        self.boundary_conditions = {
            c_s: {"left": (pybamm.Scalar(0), "Neumann"), "right": (rbc, "Neumann")}
        }

    def set_initial_conditions(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        c_init = self.phase_param.c_init
        if self.size_distribution is False:
            if self.x_average is False:
                c_s = variables[
                    f"{Domain} {phase_name}particle concentration [mol.m-3]"
                ]
            else:
                c_s = variables[
                    f"X-averaged {domain} {phase_name}particle concentration [mol.m-3]"
                ]
                c_init = pybamm.x_average(c_init)
        else:
            if self.x_average is False:
                c_s = variables[
                    f"{Domain} {phase_name}particle "
                    "concentration distribution [mol.m-3]"
                ]
                c_init = pybamm.SecondaryBroadcast(
                    c_init, f"{domain} {phase_name}particle size"
                )
            else:
                c_s = variables[
                    f"X-averaged {domain} {phase_name}particle "
                    "concentration distribution [mol.m-3]"
                ]

                c_init = pybamm.SecondaryBroadcast(
                    pybamm.x_average(c_init), f"{domain} {phase_name}particle size"
                )
        self.initial_conditions = {c_s: c_init}



#
# Base class for porosity
#
import pybamm


class BaseModel(pybamm.BaseSubModel):
    """Base class for porosity

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    """

    def __init__(self, param, options):
        super().__init__(param, options=options)

    def _get_standard_porosity_variables(self, eps_dict):
        eps = pybamm.concatenation(*eps_dict.values())

        variables = {"Porosity": eps}

        for domain, eps_k in eps_dict.items():
            Domain = domain.capitalize()
            eps_k_av = pybamm.x_average(eps_k)
            variables.update(
                {
                    f"{Domain} porosity": eps_k,
                    f"X-averaged {domain} porosity": eps_k_av,
                }
            )

        return variables

    def _get_standard_porosity_change_variables(self, depsdt_dict):
        deps_dt = pybamm.concatenation(*depsdt_dict.values())

        variables = {"Porosity change": deps_dt}
        for domain, depsdt_k in depsdt_dict.items():
            Domain = domain.capitalize()
            depsdt_k_av = pybamm.x_average(depsdt_k)
            variables.update(
                {
                    f"{Domain} porosity change [s-1]": depsdt_k,
                    f"X-averaged {domain} porosity change [s-1]": depsdt_k_av,
                }
            )

        return variables



#
# Class for leading-order pressure driven convection
#
import pybamm

from .base_through_cell_convection import BaseThroughCellModel


class Explicit(BaseThroughCellModel):
    """A submodel for the leading-order approximation of pressure-driven convection

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    """

    def __init__(self, param):
        super().__init__(param)

    def get_coupled_variables(self, variables):
        # Set up
        p_s = variables["X-averaged separator pressure [Pa]"]
        for domain in self.options.whole_cell_domains:
            if domain == "separator":
                continue
            a_j_k_av = variables[
                f"X-averaged {domain} volumetric interfacial current density [A.m-3]"
            ]
            if domain == "negative electrode":
                x_n = pybamm.standard_spatial_vars.x_n
                DeltaV_k = self.param.n.DeltaV
                p_k = (
                    -DeltaV_k
                    * a_j_k_av
                    / self.param.F
                    * (-(x_n**2) + self.param.n.L**2)
                    / 2
                    + p_s
                )
                v_box_k = -DeltaV_k * a_j_k_av / self.param.F * x_n
            elif domain == "positive electrode":
                x_p = pybamm.standard_spatial_vars.x_p
                DeltaV_k = self.param.p.DeltaV
                p_k = (
                    -DeltaV_k
                    * a_j_k_av
                    / self.param.F
                    * ((x_p - 1) ** 2 - self.param.p.L**2)
                    / 2
                    + p_s
                )
                v_box_k = -DeltaV_k * a_j_k_av / self.param.F * (x_p - self.param.L_x)
            div_v_box_k = pybamm.PrimaryBroadcast(
                -DeltaV_k * a_j_k_av / self.param.F, domain
            )

            variables.update(
                self._get_standard_convection_variables(
                    domain, v_box_k, div_v_box_k, p_k
                )
            )

        # Transverse velocity in the separator determines through-cell velocity
        div_Vbox_s = variables[
            "X-averaged separator transverse volume-averaged acceleration [m.s-2]"
        ]
        i_boundary_cc = variables["Current collector current density [A.m-2]"]
        v_box_n_right = -self.param.n.DeltaV * i_boundary_cc / self.param.F
        div_v_box_s_av = -div_Vbox_s
        div_v_box_s = pybamm.PrimaryBroadcast(div_v_box_s_av, "separator")

        # Simple formula for velocity in the separator
        x_s = pybamm.standard_spatial_vars.x_s
        v_box_s = div_v_box_s_av * (x_s - self.param.n.L) + v_box_n_right

        variables.update(
            self._get_standard_sep_velocity_variables(v_box_s, div_v_box_s)
        )
        variables.update(self._get_standard_whole_cell_velocity_variables(variables))
        variables.update(
            self._get_standard_whole_cell_acceleration_variables(variables)
        )
        variables.update(self._get_standard_whole_cell_pressure_variables(variables))

        return variables



#
# Base class for open-circuit potential with hysteresis
#
import pybamm

from . import BaseOpenCircuitPotential


class BaseHysteresisOpenCircuitPotential(BaseOpenCircuitPotential):
    """
    Base class for open-circuit potentials with single-state hysteresis

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain to implement the model, either: 'Negative' or 'Positive'.
    reaction : str
        The name of the reaction being implemented
    options: dict
        A dictionary of options to be passed to the model. See
        :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    x_average : bool
        Whether the particle concentration is averaged over the x-direction. Default is False.
    """

    def __init__(
        self, param, domain, reaction, options, phase="primary", x_average=False
    ):
        super().__init__(
            param, domain, reaction, options=options, phase=phase, x_average=x_average
        )

    def _get_hysteresis_state_variables(self):
        domain, Domain = self.domain_Domain
        domain_options = getattr(self.options, domain)
        phase_name = self.phase_name

        if self.x_average is False:
            h = pybamm.Variable(
                f"{Domain} electrode {phase_name}hysteresis state",
                domains={
                    "primary": f"{domain} electrode",
                    "secondary": "current collector",
                },
            )
            h_x_av = pybamm.x_average(h)
        else:
            h_x_av = pybamm.Variable(
                f"X-averaged {domain} electrode {phase_name}hysteresis state",
                domains={
                    "primary": "current collector",
                },
            )
            h = pybamm.PrimaryBroadcast(h_x_av, [f"{domain} electrode"])
        variables = {
            f"{Domain} electrode {phase_name}hysteresis state": h,
            f"X-averaged {domain} electrode {phase_name}hysteresis state": h_x_av,
        }
        if domain_options["particle size"] == "distribution":
            h_dist = pybamm.PrimaryBroadcast(h, [f"{domain} {phase_name}particle size"])
            h_dist_x_av = pybamm.x_average(h_dist)
            variables.update(
                {
                    f"{Domain} electrode {phase_name}hysteresis state distribution": h_dist,
                    f"X-averaged {domain} electrode {phase_name}hysteresis state distribution": h_dist_x_av,
                }
            )
        return variables

    def _get_coupled_variables(self, variables):
        domain, Domain = self.domain_Domain
        domain_options = getattr(self.options, domain)
        phase_name = self.phase_name

        if self.reaction == "lithium-ion main":
            # Get equilibrium, delithiation and lithiation OCPs
            sto_surf, sto_bulk, T, T_bulk = self._get_stoichiometry_and_temperature(
                variables
            )
            U_eq = self.phase_param.U(sto_surf, T)
            U_eq_x_av = self.phase_param.U(sto_surf, T)
            U_lith = self.phase_param.U(sto_surf, T, "lithiation")
            U_lith_bulk = self.phase_param.U(sto_bulk, T_bulk, "lithiation")
            U_delith = self.phase_param.U(sto_surf, T, "delithiation")
            U_delith_bulk = self.phase_param.U(sto_bulk, T_bulk, "delithiation")

            H = U_lith - U_delith
            H_x_av = pybamm.x_average(H)

            variables.update(
                {
                    f"{Domain} electrode {phase_name}OCP hysteresis [V]": H,
                    f"X-averaged {domain} electrode {phase_name}OCP hysteresis [V]": H_x_av,
                    f"{Domain} electrode {phase_name}equilibrium open-circuit potential [V]": U_eq,
                    f"X-averaged {domain} electrode {phase_name}equilibrium open-circuit potential [V]": U_eq_x_av,
                }
            )

            # Get correct hysteresis state variable
            if domain_options["particle size"] == "distribution":
                h = variables[
                    f"{Domain} electrode {phase_name}hysteresis state distribution"
                ]
                h_x_av = variables[
                    f"X-averaged {domain} electrode {phase_name}hysteresis state distribution"
                ]
            else:
                h = variables[f"{Domain} electrode {phase_name}hysteresis state"]
                h_x_av = variables[
                    f"X-averaged {domain} electrode {phase_name}hysteresis state"
                ]

            # check MPM
            if (
                domain_options["particle size"] == "distribution"
                and "current collector" in sto_surf.domains["secondary"]
            ):
                ocp_surf = (1 + h_x_av) / 2 * U_delith + (1 - h_x_av) / 2 * U_lith
            else:
                ocp_surf = (1 + h) / 2 * U_delith + (1 - h) / 2 * U_lith

            # Size average
            h_s_av = pybamm.size_average(h_x_av)
            ocp_bulk = (1 + h_s_av) / 2 * U_delith_bulk + (1 - h_s_av) / 2 * U_lith_bulk
            dUdT = self.phase_param.dUdT(sto_surf)

        variables.update(self._get_standard_ocp_variables(ocp_surf, ocp_bulk, dUdT))
        return variables

    def set_initial_conditions(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        h_init = self.phase_param.h_init
        if self.x_average is False:
            h = variables[f"{Domain} electrode {phase_name}hysteresis state"]
        else:
            h = variables[f"X-averaged {domain} electrode {phase_name}hysteresis state"]
            h_init = pybamm.x_average(h_init)

        self.initial_conditions[h] = h_init



#
# Class for ambient surface temperature submodel
import pybamm


class Ambient(pybamm.BaseSubModel):
    """
    Class for setting surface temperature equal to ambient temperature.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.

    """

    def __init__(self, param, options=None):
        super().__init__(param, options=options)

    def get_coupled_variables(self, variables):
        T_amb = variables["Ambient temperature [K]"]
        T_amb_av = variables["Volume-averaged ambient temperature [K]"]

        variables.update(
            {
                "Surface temperature [K]": T_amb,
                "Volume-averaged surface temperature [K]": T_amb_av,
                "Environment total cooling [W]": pybamm.Scalar(0),
            }
        )
        return variables



def electrolyte_diffusivity_Nyman2008(c_e, T):
    """
    Diffusivity of LiPF6 in EC:EMC (3:7) as a function of ion concentration. The data
    comes from [1]

    References
    ----------
    .. [1] A. Nyman, M. Behm, and G. Lindbergh, "Electrochemical characterisation and
    modelling of the mass transport phenomena in LiPF6-EC-EMC electrolyte,"
    Electrochim. Acta, vol. 53, no. 22, pp. 6356–6365, 2008.

    Parameters
    ----------
    c_e: :class:`pybamm.Symbol`
        Dimensional electrolyte concentration
    T: :class:`pybamm.Symbol`
        Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """

    D_c_e = 8.794e-11 * (c_e / 1000) ** 2 - 3.972e-10 * (c_e / 1000) + 4.862e-10

    # Nyman et al. (2008) does not provide temperature dependence

    return D_c_e


def electrolyte_conductivity_Nyman2008(c_e, T):
    """
    Conductivity of LiPF6 in EC:EMC (3:7) as a function of ion concentration. The data
    comes from [1].

    References
    ----------
    .. [1] A. Nyman, M. Behm, and G. Lindbergh, "Electrochemical characterisation and
    modelling of the mass transport phenomena in LiPF6-EC-EMC electrolyte,"
    Electrochim. Acta, vol. 53, no. 22, pp. 6356–6365, 2008.

    Parameters
    ----------
    c_e: :class:`pybamm.Symbol`
        Dimensional electrolyte concentration
    T: :class:`pybamm.Symbol`
        Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """

    sigma_e = (
        0.1297 * (c_e / 1000) ** 3 - 2.51 * (c_e / 1000) ** 1.5 + 3.329 * (c_e / 1000)
    )

    # Nyman et al. (2008) does not provide temperature dependence

    return sigma_e


def get_parameter_values():
    """
    Example parameter values for use with MSMR models. The thermodynamic parameters
    are for Graphite and NMC622, and are taken from Table 1 of the paper

        Mark Verbrugge, Daniel Baker, Brian Koch, Xingcheng Xiao and Wentian Gu.
        Thermodynamic Model for Substitutional Materials: Application to Lithiated
        Graphite, Spinel Manganese Oxide, Iron Phosphate, and Layered
        Nickel-Manganese-Cobalt Oxide. Journal of The Electrochemical Society,
        164(11):3243-3253, 2017. doi:10.1149/2.0341708jes.

    The remaining value are based on a parameterization of the LG M50 cell, from the
    paper

        Chang-Hui Chen, Ferran Brosa Planella, Kieran O'Regan, Dominika Gastol, W.
        Dhammika Widanage, and Emma Kendrick. Development of Experimental Techniques for
        Parameterization of Multi-scale Lithium-ion Battery Models. Journal of The
        Electrochemical Society, 167(8):080534, 2020. doi:10.1149/1945-7111/ab9050.

    and references therein. Verbrugge et al. (2017) does not provide kinetic parameters
    so we set the reference exchange current density to 5 A.m-2 for the positive
    electrode reactions and 2.7 A.m-2 for the negative electrode reactions, which are
    the values used in the Chen et al. (2020) paper. We also assume that the
    exchange-current density is symmetric. Note: the 4th reaction in the positive
    electrode gave unphysical results so we set the reference exchange current density
    and symmetry factor to 1e6 and 1, respectively. The parameter values are intended
    to serve as an example set to use with the MSMR model and do not claim to match any
    experimental cycling data.
    """
    return {
        "chemistry": "lithium_ion",
        # cell
        "Negative electrode thickness [m]": 8.52e-05,
        "Separator thickness [m]": 1.2e-05,
        "Positive electrode thickness [m]": 7.56e-05,
        "Electrode height [m]": 0.065,
        "Electrode width [m]": 1.58,
        "Nominal cell capacity [A.h]": 5.0,
        "Current function [A]": 5.0,
        "Contact resistance [Ohm]": 0,
        # negative electrode
        "Number of reactions in negative electrode": 6,
        "Negative electrode host site standard potential (0) [V]": 0.08843,
        "Negative electrode host site occupancy fraction (0)": 0.43336,
        "Negative electrode host site ideality factor (0)": 0.08611,
        "Negative electrode host site charge transfer coefficient (0)": 0.5,
        "Negative electrode host site reference exchange-current density (0) [A.m-2]": 2.7,
        "Negative electrode host site standard potential (1) [V]": 0.12799,
        "Negative electrode host site occupancy fraction (1)": 0.23963,
        "Negative electrode host site ideality factor (1)": 0.08009,
        "Negative electrode host site charge transfer coefficient (1)": 0.5,
        "Negative electrode host site reference exchange-current density (1) [A.m-2]": 2.7,
        "Negative electrode host site standard potential (2) [V]": 0.14331,
        "Negative electrode host site occupancy fraction (2)": 0.15018,
        "Negative electrode host site ideality factor (2)": 0.72469,
        "Negative electrode host site charge transfer coefficient (2)": 0.5,
        "Negative electrode host site reference exchange-current density (2) [A.m-2]": 2.7,
        "Negative electrode host site standard potential (3) [V]": 0.16984,
        "Negative electrode host site occupancy fraction (3)": 0.05462,
        "Negative electrode host site ideality factor (3)": 2.53277,
        "Negative electrode host site charge transfer coefficient (3)": 0.5,
        "Negative electrode host site reference exchange-current density (3) [A.m-2]": 2.7,
        "Negative electrode host site standard potential (4) [V]": 0.21446,
        "Negative electrode host site occupancy fraction (4)": 0.06744,
        "Negative electrode host site ideality factor (4)": 0.09470,
        "Negative electrode host site charge transfer coefficient (4)": 0.5,
        "Negative electrode host site reference exchange-current density (4) [A.m-2]": 2.7,
        "Negative electrode host site standard potential (5) [V]": 0.36325,
        "Negative electrode host site occupancy fraction (5)": 0.05476,
        "Negative electrode host site ideality factor (5)": 5.97354,
        "Negative electrode host site charge transfer coefficient (5)": 0.5,
        "Negative electrode host site reference exchange-current density (5) [A.m-2]": 2.7,
        "Negative electrode conductivity [S.m-1]": 215.0,
        "Maximum concentration in negative electrode [mol.m-3]": 33133.0,
        "Negative particle diffusivity [m2.s-1]": 3.3e-14,
        "Negative electrode porosity": 0.25,
        "Negative electrode active material volume fraction": 0.75,
        "Negative particle radius [m]": 5.86e-06,
        "Negative electrode Bruggeman coefficient (electrolyte)": 1.5,
        "Negative electrode Bruggeman coefficient (electrode)": 0,
        "Negative electrode OCP entropic change [V.K-1]": 0.0,
        # positive electrode
        "Number of reactions in positive electrode": 4,
        "Positive electrode host site standard potential (0) [V]": 3.62274,
        "Positive electrode host site occupancy fraction (0)": 0.13442,
        "Positive electrode host site ideality factor (0)": 0.96710,
        "Positive electrode host site charge transfer coefficient (0)": 0.5,
        "Positive electrode host site reference exchange-current density (0) [A.m-2]": 5,
        "Positive electrode host site standard potential (1) [V]": 3.72645,
        "Positive electrode host site occupancy fraction (1)": 0.32460,
        "Positive electrode host site ideality factor (1)": 1.39712,
        "Positive electrode host site charge transfer coefficient (1)": 0.5,
        "Positive electrode host site reference exchange-current density (1) [A.m-2]": 5,
        "Positive electrode host site standard potential (2) [V]": 3.90575,
        "Positive electrode host site occupancy fraction (2)": 0.21118,
        "Positive electrode host site ideality factor (2)": 3.50500,
        "Positive electrode host site charge transfer coefficient (2)": 0.5,
        "Positive electrode host site reference exchange-current density (2) [A.m-2]": 5,
        "Positive electrode host site standard potential (3) [V]": 4.22955,
        "Positive electrode host site occupancy fraction (3)": 0.32980,
        "Positive electrode host site ideality factor (3)": 5.52757,
        "Positive electrode host site charge transfer coefficient (3)": 1,
        "Positive electrode host site reference exchange-current density (3) [A.m-2]": 1e6,
        "Positive electrode conductivity [S.m-1]": 0.18,
        "Maximum concentration in positive electrode [mol.m-3]": 63104.0,
        "Positive particle diffusivity [m2.s-1]": 4e-15,
        "Positive electrode porosity": 0.335,
        "Positive electrode active material volume fraction": 0.665,
        "Positive particle radius [m]": 5.22e-06,
        "Positive electrode Bruggeman coefficient (electrolyte)": 1.5,
        "Positive electrode Bruggeman coefficient (electrode)": 0,
        "Positive electrode OCP entropic change [V.K-1]": 0.0,
        # separator
        "Separator porosity": 0.47,
        "Separator Bruggeman coefficient (electrolyte)": 1.5,
        # electrolyte
        "Initial concentration in electrolyte [mol.m-3]": 1000.0,
        "Cation transference number": 0.2594,
        "Thermodynamic factor": 1.0,
        "Electrolyte diffusivity [m2.s-1]": electrolyte_diffusivity_Nyman2008,
        "Electrolyte conductivity [S.m-1]": electrolyte_conductivity_Nyman2008,
        # experiment
        "Reference temperature [K]": 298.15,
        "Total heat transfer coefficient [W.m-2.K-1]": 10.0,
        "Ambient temperature [K]": 298.15,
        "Number of electrodes connected in parallel to make a cell": 1.0,
        "Number of cells connected in series to make a battery": 1.0,
        "Lower voltage cut-off [V]": 2.8,
        "Upper voltage cut-off [V]": 4.2,
        "Open-circuit voltage at 0% SOC [V]": 2.8,
        "Open-circuit voltage at 100% SOC [V]": 4.2,
        "Initial temperature [K]": 298.15,
        "Initial voltage in negative electrode [V]": 0.01,
        "Initial voltage in positive electrode [V]": 4.19,
        # citations
        "citations": ["Verbrugge2017", "Baker2018", "Chen2020"],
    }



#
# Class for composite surface form electrolyte conductivity employing stefan-maxwell
#
import pybamm
from pybamm.models.submodels.electrolyte_conductivity.composite_conductivity import (
    Composite,
)


class BaseModel(Composite):
    """
    Base class for composite conservation of charge in the electrolyte employing
    the Stefan-Maxwell constitutive equations employing the surface potential difference
    formulation.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain in which the model holds
    options : dict
        Additional options to pass to the model
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options)

    def get_fundamental_variables(self):
        if self.domain == "separator":
            return {}

        delta_phi_av = pybamm.Variable(
            f"X-averaged {self.domain} electrode surface potential difference [V]",
            domain="current collector",
            reference=self.domain_param.prim.U_init,
        )

        variables = self._get_standard_average_surface_potential_difference_variables(
            delta_phi_av
        )
        return variables

    def get_coupled_variables(self, variables):
        Domain = self.domain.capitalize()
        # Only update coupled variables once
        if self.domain == "negative":
            variables.update(super().get_coupled_variables(variables))

        phi_s = variables[f"{Domain} electrode potential [V]"]
        phi_e = variables[f"{Domain} electrolyte potential [V]"]
        delta_phi = phi_s - phi_e
        variables.update(
            self._get_standard_surface_potential_difference_variables(delta_phi)
        )
        return variables

    def set_initial_conditions(self, variables):
        domain = self.domain

        delta_phi = variables[
            f"X-averaged {domain} electrode surface potential difference [V]"
        ]
        delta_phi_init = self.domain_param.prim.U_init

        self.initial_conditions = {delta_phi: delta_phi_init}

    def set_boundary_conditions(self, variables):
        if self.domain == "negative":
            phi_e = variables["Electrolyte potential [V]"]
            self.boundary_conditions = {
                phi_e: {
                    "left": (pybamm.Scalar(0), "Neumann"),
                    "right": (pybamm.Scalar(0), "Neumann"),
                }
            }


class CompositeDifferential(BaseModel):
    """
    Composite model for conservation of charge in the electrolyte employing the
    Stefan-Maxwell constitutive equations employing the surface potential difference
    formulation and where capacitance is present.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain in which the model holds
    options : dict
        Additional options to pass to the model
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options)

    def set_rhs(self, variables):
        domain = self.domain

        a = variables[
            f"X-averaged {domain} electrode surface area to volume ratio [m-1]"
        ]

        sum_a_j = variables[
            f"Sum of x-averaged {domain} electrode volumetric "
            "interfacial current densities [A.m-3]"
        ]

        sum_a_j_av = variables[
            f"X-averaged {domain} electrode total volumetric "
            "interfacial current density [A.m-3]"
        ]
        delta_phi = variables[
            f"X-averaged {domain} electrode surface potential difference [V]"
        ]

        T = variables[f"X-averaged {domain} electrode temperature [K]"]

        C_dl = self.domain_param.C_dl(T)

        self.rhs[delta_phi] = 1 / (a * C_dl) * (sum_a_j_av - sum_a_j)


class CompositeAlgebraic(BaseModel):
    """
    Composite model for conservation of charge in the electrolyte employing the
    Stefan-Maxwell constitutive equations employing the surface potential difference
    formulation.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain in which the model holds
    options : dict
        Additional options to pass to the model
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options)

    def set_algebraic(self, variables):
        domain = self.domain

        sum_a_j = variables[
            f"Sum of x-averaged {domain} electrode volumetric "
            "interfacial current densities [A.m-3]"
        ]

        sum_a_j_av = variables[
            f"X-averaged {domain} electrode total volumetric "
            "interfacial current density [A.m-3]"
        ]
        delta_phi = variables[
            f"X-averaged {domain} electrode surface potential difference [V]"
        ]

        self.algebraic[delta_phi] = (sum_a_j_av - sum_a_j) / self.param.a_j_scale



#
# Class for ohmic electrodes in the surface potential formulation
#
import pybamm

from .base_ohm import BaseModel


class SurfaceForm(BaseModel):
    """A submodel for the electrode with Ohm's law in the surface potential
    formulation.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        Either 'negative' or 'positive'
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options=options)

    def get_coupled_variables(self, variables):
        Domain = self.domain.capitalize()

        x_n = pybamm.standard_spatial_vars.x_n
        x_p = pybamm.standard_spatial_vars.x_p
        i_boundary_cc = variables["Current collector current density [A.m-2]"]
        i_e = variables[f"{Domain} electrolyte current density [A.m-2]"]
        tor = variables[f"{Domain} electrode transport efficiency"]
        phi_s_cn = variables["Negative current collector potential [V]"]
        T = variables[f"{Domain} electrode temperature [K]"]

        conductivity = self.domain_param.sigma(T) * tor
        i_s = i_boundary_cc - i_e

        if self.domain == "negative":
            phi_s = phi_s_cn - pybamm.IndefiniteIntegral(i_s / conductivity, x_n)

        elif self.domain == "positive":
            phi_e_s = variables["Separator electrolyte potential [V]"]
            delta_phi_p = variables[
                "Positive electrode surface potential difference [V]"
            ]

            phi_s = -pybamm.IndefiniteIntegral(i_s / conductivity, x_p) + (
                pybamm.boundary_value(phi_e_s, "right")
                + pybamm.boundary_value(delta_phi_p, "left")
            )

        variables.update(self._get_standard_potential_variables(phi_s))
        variables.update(self._get_standard_current_variables(i_s))

        if (
            self.options.electrode_types["negative"] == "planar"
            or "Negative electrode current density [A.m-2]" in variables
        ) and "Positive electrode current density [A.m-2]" in variables:
            variables.update(self._get_standard_whole_cell_variables(variables))

        return variables



#
# Exception classes
#


class DomainError(Exception):
    """Domain error: an operation was attempted on nodes with un-matched domains."""

    pass


class OptionError(Exception):
    """Option error: an unknown option was given."""

    pass


class OptionWarning(UserWarning):
    """Option warning: the chosen options may not give the desired output."""

    pass


class GeometryError(Exception):
    """Geometry error: Raised if the an unimplemented geometry is used."""

    pass


class ModelError(Exception):
    """Model error: the model is not well-posed (can be before or after processing)"""

    pass


class SolverError(Exception):
    """
    Solver error: a solution to the model could not be found with the chosen settings
    """

    def __init__(self, *args):
        self.message = args[0]


class SolverWarning(UserWarning):
    """Solver warning: the chosen solver settings may not give the desired output."""

    pass


class ShapeError(Exception):
    """Shape error: cannot evaluate an object to find its shape."""

    pass


class ModelWarning(UserWarning):
    """
    Model warning: the model does not contain all of the standard output variables
    """

    pass


class DiscretisationError(Exception):
    """A model could not be discretised."""

    pass


class InvalidModelJSONError(Exception):
    """Raised when a model JSON file is invalid or cannot be parsed."""

    pass



import pybamm

from .util import _has_hysteresis, check_if_composite


def _set_hysteresis_branch(parameter_values, electrode, direction, options, phase=None):
    phase = phase or ""
    if phase != "":
        phase_prefactor = phase.capitalize() + ": "
    else:
        phase_prefactor = ""
    if direction is None:
        initial_hysteresis_branch = 0
    else:
        if (direction == "discharge" and electrode == "negative") or (
            direction == "charge" and electrode == "positive"
        ):
            initial_hysteresis_branch = 1
        elif (direction == "charge" and electrode == "negative") or (
            direction == "discharge" and electrode == "positive"
        ):
            initial_hysteresis_branch = -1
        else:
            raise ValueError(f"Invalid direction: {direction}")
    parameter_values.update(
        {
            f"{phase_prefactor}Initial hysteresis state in {electrode} electrode": initial_hysteresis_branch,
        }
    )
    return parameter_values


def set_initial_state(
    initial_value,
    parameter_values,
    direction=None,
    param=None,
    known_value="cyclable lithium capacity",
    inplace=True,
    options=None,
    inputs=None,
    tol=1e-6,
):
    """
    Set the values of the parameters representing the initial model states.

    Parameters
    ----------
    initial_value : float
        Target initial value.
        If float, interpreted as SOC, must be between 0 and 1.
        If string e.g. "4 V", interpreted as voltage, must be between V_min and V_max.
    parameter_values : :class:`pybamm.ParameterValues`
        Parameters and their corresponding values.
    param : :class:`pybamm.LithiumIonParameters`, optional
        The symbolic parameter set to use for the simulation.
        If not provided, the default parameter set will be used.
    known_value : str, optional
        The known value needed to complete the electrode SOH model.
        Can be "cyclable lithium capacity" (default) or "cell capacity".
    inplace: bool, optional
        If True, replace the parameters values in place. Otherwise, return a new set of
        parameter values. Default is True.
    options : dict-like, optional
        A dictionary of options to be passed to the model, see
        :class:`pybamm.BatteryModelOptions`.
    inputs : dict, optional
        A dictionary of input parameters to pass to the model when solving.
    tol : float, optional
        The tolerance for the solver used to compute the initial stoichiometries.
        A lower value results in higher precision but may increase computation time.
        Default is 1e-6.
    """
    options = options or {}
    parameter_values = parameter_values if inplace else parameter_values.copy()
    param = param or pybamm.LithiumIonParameters(options)

    for electrode in ["negative", "positive"]:
        if check_if_composite(options, electrode):
            for phase in ["primary", "secondary"]:
                if _has_hysteresis(electrode, options, phase):
                    parameter_values = _set_hysteresis_branch(
                        parameter_values, electrode, direction, options, phase
                    )
        else:
            if _has_hysteresis(electrode, options):
                parameter_values = _set_hysteresis_branch(
                    parameter_values, electrode, direction, options
                )

    if options is not None and options.get("open-circuit potential", None) == "MSMR":
        """
        Set the initial OCP of each electrode, based on the initial SOC or voltage.
        """
        Un, Up = pybamm.lithium_ion.get_initial_ocps(
            initial_value,
            parameter_values,
            direction=direction,
            param=param,
            known_value=known_value,
            options=options,
            tol=tol,
            inputs=inputs,
        )
        parameter_values.update(
            {
                "Initial voltage in negative electrode [V]": Un,
                "Initial voltage in positive electrode [V]": Up,
            }
        )
    elif options is not None and options.get("working electrode", None) == "positive":
        """
        Set the initial stoichiometry of the working electrode, based on the initial
        SOC or voltage.
        """
        results = pybamm.lithium_ion.get_initial_stoichiometry_half_cell(
            initial_value,
            parameter_values,
            param=param,
            options=options,
            tol=tol,
            inputs=inputs,
            direction=direction,
        )
        _set_concentration_from_stoich(
            parameter_values,
            param,
            "positive",
            "primary",
            results["x"],
            inputs,
            options,
        )
        if check_if_composite(options, "positive"):
            _set_concentration_from_stoich(
                parameter_values,
                param,
                "positive",
                "secondary",
                results["x_2"],
                inputs,
                options,
            )
    elif options is not None and (
        check_if_composite(options, "positive")
        or check_if_composite(options, "negative")
    ):
        """
        Set the initial stoichiometry of each electrode, based on the initial SOC or
        voltage.
        """
        initial_stoichs = pybamm.lithium_ion.get_initial_stoichiometries_composite(
            initial_value,
            parameter_values,
            direction=direction,
            param=param,
            options=options,
            tol=tol,
            inputs=inputs,
            known_value=known_value,
        )
        _set_concentration_from_stoich(
            parameter_values,
            param,
            "positive",
            "primary",
            initial_stoichs["y_init_1"],
            inputs,
            options,
        )
        _set_concentration_from_stoich(
            parameter_values,
            param,
            "negative",
            "primary",
            initial_stoichs["x_init_1"],
            inputs,
            options,
        )
        if check_if_composite(options, "positive"):
            _set_concentration_from_stoich(
                parameter_values,
                param,
                "positive",
                "secondary",
                initial_stoichs["y_init_2"],
                inputs,
                options,
            )
        if check_if_composite(options, "negative"):
            _set_concentration_from_stoich(
                parameter_values,
                param,
                "negative",
                "secondary",
                initial_stoichs["x_init_2"],
                inputs,
                options,
            )
    else:
        """
        Set the initial stoichiometry of each electrode, based on the initial SOC or
        voltage.
        """
        x, y = pybamm.lithium_ion.get_initial_stoichiometries(
            initial_value,
            parameter_values,
            direction=direction,
            param=param,
            known_value=known_value,
            options=options,
            tol=tol,
            inputs=inputs,
        )
        _set_concentration_from_stoich(
            parameter_values, param, "negative", "primary", x, inputs, options
        )
        _set_concentration_from_stoich(
            parameter_values, param, "positive", "primary", y, inputs, options
        )

    return parameter_values


def _set_concentration_from_stoich(
    parameter_values, param, electrode, phase, stoich, inputs, options
):
    if electrode == "positive":
        electrode_param = param.p
    else:
        electrode_param = param.n
    if phase == "primary":
        phase_param = electrode_param.prim
    elif phase == "secondary":
        phase_param = electrode_param.sec
    else:
        raise ValueError(f"Invalid phase: {phase}")
    if check_if_composite(options, electrode):
        phase_prefactor = phase.capitalize() + ": "
    else:
        phase_prefactor = ""
    parameter_values.update(
        {
            f"{phase_prefactor}Initial concentration in {electrode} electrode [mol.m-3]": stoich
            * parameter_values.evaluate(phase_param.c_max, inputs=inputs)
        }
    )



#
# Class for explicit surface form potentials
#
import pybamm
from pybamm.models.submodels.electrolyte_conductivity.base_electrolyte_conductivity import (
    BaseElectrolyteConductivity,
)


class Explicit(BaseElectrolyteConductivity):
    """Class for deriving surface potential difference variables from the electrode
    and electrolyte potentials

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain in which the model holds
    options : dict
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options):
        super().__init__(param, domain, options)

    def get_coupled_variables(self, variables):
        # skip for separator
        if self.domain == "separator":
            return variables

        Domain = self.domain.capitalize()
        phi_s = variables[f"{Domain} electrode potential [V]"]
        phi_e = variables[f"{Domain} electrolyte potential [V]"]
        delta_phi = phi_s - phi_e
        variables.update(
            self._get_standard_surface_potential_difference_variables(delta_phi)
        )

        delta_phi_av = pybamm.x_average(delta_phi)
        variables.update(
            self._get_standard_average_surface_potential_difference_variables(
                delta_phi_av
            )
        )
        return variables

    def set_boundary_conditions(self, variables):
        # don't set any boundary conditions
        return



from .x_full import OneDimensionalX
from .pouch_cell_1D_current_collectors import CurrentCollector1D
from .pouch_cell_2D_current_collectors import CurrentCollector2D

__all__ = ['pouch_cell_1D_current_collectors',
           'pouch_cell_2D_current_collectors', 'x_full']



#
# Submodel for no convection in transverse directions
#
import pybamm

from .base_transverse_convection import BaseTransverseModel


class NoConvection(BaseTransverseModel):
    """
    Submodel for no convection in transverse directions

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, options=None):
        super().__init__(param, options=options)

    def get_fundamental_variables(self):
        p_s = pybamm.PrimaryBroadcast(0, "current collector")
        variables = self._get_standard_separator_pressure_variables(p_s)

        Vbox_s = pybamm.PrimaryBroadcast(0, "current collector")
        variables.update(
            self._get_standard_transverse_velocity_variables(Vbox_s, "velocity")
        )

        div_Vbox_s = pybamm.PrimaryBroadcast(0, "current collector")

        variables.update(
            self._get_standard_transverse_velocity_variables(div_Vbox_s, "acceleration")
        )

        return variables



#
# Convert a PyBaMM expression tree to a CasADi expression tree
#
from __future__ import annotations

import casadi
import numpy as np
from scipy import interpolate, special

import pybamm


class CasadiConverter:
    def __init__(self, casadi_symbols=None):
        self._casadi_symbols = casadi_symbols or {}

        pybamm.citations.register("Andersson2019")

    def convert(
        self,
        symbol: pybamm.Symbol,
        t: casadi.MX,
        y: casadi.MX,
        y_dot: casadi.MX,
        inputs: dict | None,
    ) -> casadi.MX:
        """
        This function recurses down the tree, converting the PyBaMM expression tree to
        a CasADi expression tree

        Parameters
        ----------
        symbol : :class:`pybamm.Symbol`
            The symbol to convert
        t : :class:`casadi.MX`
            A casadi symbol representing time
        y : :class:`casadi.MX`
            A casadi symbol representing state vectors
        y_dot : :class:`casadi.MX`
            A casadi symbol representing time derivatives of state vectors
        inputs : dict
            A dictionary of casadi symbols representing parameters

        Returns
        -------
        :class:`casadi.MX`
            The converted symbol
        """
        try:
            return self._casadi_symbols[symbol]
        except KeyError:
            # Change inputs to empty dictionary if it's None
            inputs = inputs or {}
            casadi_symbol = self._convert(symbol, t, y, y_dot, inputs)
            self._casadi_symbols[symbol] = casadi_symbol

            return casadi_symbol

    def _convert(self, symbol, t, y, y_dot, inputs):
        """See :meth:`CasadiConverter.convert()`."""
        if isinstance(
            symbol,
            pybamm.Scalar | pybamm.Array | pybamm.Time | pybamm.InputParameter,
        ):
            return casadi.MX(symbol.evaluate(t, y, y_dot, inputs))

        elif isinstance(symbol, pybamm.StateVector):
            if y is None:
                raise ValueError("Must provide a 'y' for converting state vectors")
            return casadi.vertcat(*[y[y_slice] for y_slice in symbol.y_slices])

        elif isinstance(symbol, pybamm.StateVectorDot):
            if y_dot is None:
                raise ValueError("Must provide a 'y_dot' for converting state vectors")
            return casadi.vertcat(*[y_dot[y_slice] for y_slice in symbol.y_slices])

        elif isinstance(symbol, pybamm.BinaryOperator):
            left, right = symbol.children
            # process children
            converted_left = self.convert(left, t, y, y_dot, inputs)
            converted_right = self.convert(right, t, y, y_dot, inputs)

            if isinstance(symbol, pybamm.Modulo):
                return casadi.fmod(converted_left, converted_right)
            if isinstance(symbol, pybamm.Minimum):
                return casadi.fmin(converted_left, converted_right)
            if isinstance(symbol, pybamm.Maximum):
                return casadi.fmax(converted_left, converted_right)
            if isinstance(symbol, pybamm.KroneckerProduct):
                return casadi.kron(converted_left, converted_right)

            # _binary_evaluate defined in derived classes for specific rules
            return symbol._binary_evaluate(converted_left, converted_right)

        elif isinstance(symbol, pybamm.UnaryOperator):
            converted_child = self.convert(symbol.child, t, y, y_dot, inputs)
            if isinstance(symbol, pybamm.Transpose):
                return converted_child.T
            if isinstance(symbol, pybamm.AbsoluteValue):
                return casadi.fabs(converted_child)
            if isinstance(symbol, pybamm.Floor):
                return casadi.floor(converted_child)
            if isinstance(symbol, pybamm.Ceiling):
                return casadi.ceil(converted_child)
            return symbol._unary_evaluate(converted_child)

        elif isinstance(symbol, pybamm.Function):
            converted_children = [
                self.convert(child, t, y, y_dot, inputs) for child in symbol.children
            ]
            # Special functions
            if symbol.function == np.min:
                return casadi.mmin(*converted_children)
            elif symbol.function == np.max:
                return casadi.mmax(*converted_children)
            elif symbol.function == np.abs:
                return casadi.fabs(*converted_children)
            elif symbol.function == np.sqrt:
                return casadi.sqrt(*converted_children)
            elif symbol.function == np.sin:
                return casadi.sin(*converted_children)
            elif symbol.function == np.arcsinh:
                return casadi.arcsinh(*converted_children)
            elif symbol.function == np.arccosh:
                return casadi.arccosh(*converted_children)
            elif symbol.function == np.tanh:
                return casadi.tanh(*converted_children)
            elif symbol.function == np.cosh:
                return casadi.cosh(*converted_children)
            elif symbol.function == np.sinh:
                return casadi.sinh(*converted_children)
            elif symbol.function == np.cos:
                return casadi.cos(*converted_children)
            elif symbol.function == np.exp:
                return casadi.exp(*converted_children)
            elif symbol.function == np.log:
                return casadi.log(*converted_children)
            elif symbol.function == np.sign:
                return casadi.sign(*converted_children)
            elif symbol.function == special.erf:
                return casadi.erf(*converted_children)
            elif isinstance(symbol, pybamm.Interpolant):
                if symbol.interpolator == "linear":
                    solver = "linear"
                elif symbol.interpolator == "cubic":
                    solver = "bspline"
                elif symbol.interpolator == "pchip":
                    x_np = np.array(symbol.x[0])
                    y_np = np.array(symbol.y)
                    pchip_interp = interpolate.PchipInterpolator(x_np, y_np)
                    d_np = pchip_interp.derivative()(x_np)
                    x = converted_children[0]

                    def hermite_poly(i):
                        x0 = x_np[i]
                        x1 = x_np[i + 1]
                        h_val = x1 - x0
                        h_val_mx = casadi.MX(h_val)
                        y0 = casadi.MX(y_np[i])
                        y1 = casadi.MX(y_np[i + 1])
                        d0 = casadi.MX(d_np[i])
                        d1 = casadi.MX(d_np[i + 1])
                        xn = (x - x0) / h_val_mx
                        h00 = 2 * xn**3 - 3 * xn**2 + 1
                        h10 = xn**3 - 2 * xn**2 + xn
                        h01 = -2 * xn**3 + 3 * xn**2
                        h11 = xn**3 - xn**2
                        return (
                            h00 * y0
                            + h10 * h_val_mx * d0
                            + h01 * y1
                            + h11 * h_val_mx * d1
                        )

                    # Build piecewise polynomial for points inside the domain.
                    inside = casadi.MX.zeros(x.shape)
                    for i in range(len(x_np) - 1):
                        cond = casadi.logic_and(x >= x_np[i], x <= x_np[i + 1])
                        inside = casadi.if_else(cond, hermite_poly(i), inside)

                    # Extrapolation:
                    left = hermite_poly(0)  # For x < x_np[0]
                    right = hermite_poly(len(x_np) - 2)  # For x > x_np[-1]

                    # if greater than the maximum, use right; otherwise, use the piecewise value.
                    result = casadi.if_else(
                        x < x_np[0], left, casadi.if_else(x > x_np[-1], right, inside)
                    )
                    return result
                else:  # pragma: no cover
                    raise NotImplementedError(
                        f"Unknown interpolator: {symbol.interpolator}"
                    )

                if len(converted_children) == 1:
                    if solver == "linear":
                        test = casadi.MX.interpn_linear(
                            symbol.x, symbol.y.flatten(), converted_children
                        )
                        if test.shape[0] == 1 and test.shape[1] > 1:
                            # for some reason, pybamm.Interpolant always returns a column vector, so match that
                            test = test.T
                        return test
                    elif solver == "bspline":
                        bspline = interpolate.make_interp_spline(
                            symbol.x[0], symbol.y, k=3
                        )
                        knots = [bspline.t]
                        coeffs = bspline.c.flatten()
                        degree = [bspline.k]
                        m = len(coeffs) // len(symbol.x[0])
                        f = casadi.Function.bspline(
                            symbol.name, knots, coeffs, degree, m
                        )
                        return f(converted_children[0])
                    else:
                        return casadi.interpolant(
                            "LUT", solver, symbol.x, symbol.y.flatten()
                        )(*converted_children)
                elif len(converted_children) in [2, 3]:
                    if solver == "linear":
                        return casadi.MX.interpn_linear(
                            symbol.x,
                            symbol.y.ravel(order="F"),
                            converted_children,
                        )
                    elif solver == "bspline" and len(converted_children) == 2:
                        bspline = interpolate.RectBivariateSpline(
                            symbol.x[0], symbol.x[1], symbol.y
                        )
                        [tx, ty, c] = bspline.tck
                        [kx, ky] = bspline.degrees
                        knots = [tx, ty]
                        coeffs = c
                        degree = [kx, ky]
                        m = 1
                        f = casadi.Function.bspline(
                            symbol.name, knots, coeffs, degree, m
                        )
                        return f(casadi.hcat(converted_children).T).T
                    else:
                        LUT = casadi.interpolant(
                            "LUT", solver, symbol.x, symbol.y.ravel(order="F")
                        )
                        res = LUT(casadi.hcat(converted_children).T).T
                        return res
                else:  # pragma: no cover
                    raise ValueError(
                        f"Invalid converted_children count: {len(converted_children)}"
                    )

            elif symbol.function.__name__.startswith("elementwise_grad_of_"):
                differentiating_child_idx = int(symbol.function.__name__[-1])
                # Create dummy symbolic variables in order to differentiate using CasADi
                dummy_vars = [
                    casadi.MX.sym("y_" + str(i)) for i in range(len(converted_children))
                ]
                func_diff = casadi.gradient(
                    symbol.differentiated_function(*dummy_vars),
                    dummy_vars[differentiating_child_idx],
                )
                # Create function and evaluate it using the children
                casadi_func_diff = casadi.Function("func_diff", dummy_vars, [func_diff])
                return casadi_func_diff(*converted_children)
            # Other functions
            else:
                return symbol._function_evaluate(converted_children)
        elif isinstance(symbol, pybamm.Concatenation):
            converted_children = [
                self.convert(child, t, y, y_dot, inputs) for child in symbol.children
            ]
            if isinstance(symbol, pybamm.NumpyConcatenation | pybamm.SparseStack):
                return casadi.vertcat(*converted_children)
            # DomainConcatenation specifies a particular ordering for the concatenation,
            # which we must follow
            elif isinstance(symbol, pybamm.DomainConcatenation):
                slice_starts = []
                all_child_vectors = []
                for i in range(symbol.secondary_dimensions_npts):
                    child_vectors = []
                    for child_var, slices in zip(
                        converted_children, symbol._children_slices, strict=True
                    ):
                        for child_dom, child_slice in slices.items():
                            slice_starts.append(symbol._slices[child_dom][i].start)
                            child_vectors.append(
                                child_var[child_slice[i].start : child_slice[i].stop]
                            )
                    all_child_vectors.extend(
                        [
                            v
                            for _, v in sorted(
                                zip(slice_starts, child_vectors, strict=False)
                            )
                        ]
                    )
                return casadi.vertcat(*all_child_vectors)

        else:
            raise TypeError(
                f"Cannot convert symbol of type '{type(symbol)}' to CasADi. Symbols must all be "
                "'linear algebra' at this stage."
            )



from .base_ohm import BaseModel
from .composite_ohm import Composite
from .full_ohm import Full
from .leading_ohm import LeadingOrder
from .surface_form_ohm import SurfaceForm
from .li_metal import LithiumMetalExplicit, LithiumMetalSurfaceForm

__all__ = ['base_ohm', 'composite_ohm', 'full_ohm', 'leading_ohm', 'li_metal',
           'surface_form_ohm']
